"""RESQML grid module handling grid pixel maps."""

import numpy as np
import resqpy.olio.point_inclusion as pip


def pixel_maps(grid, origin, width, height, dx, dy = None, k0 = None, vertical_ref = 'top'):
    """Makes a mapping from pixels to cell j, i indices, based on split horizon points for a single horizon.

    args:
       origin (float pair): x, y of south west corner of area covered by pixel rectangle, in local crs
       width (int): the width of the pixel rectangle (number of pixels)
       height (int): the height of the pixel rectangle (number of pixels)
       dx (float): the size (west to east) of a pixel, in locel crs
       dy (float, optional): the size (south to north) of a pixel, in locel crs; defaults to dx
       k0 (int, default None): if present, the single layer to create a 2D pixel map for; if None, a 3D map
          is created with one layer per layer of the grid
       vertical_ref (string, default 'top'): 'top' or 'base'

    returns:
       numpy int array of shape (height, width, 2), or (nk, height, width, 2), being the j, i indices of cells
       that the pixel centres lie within; values of -1 are used as null (ie. pixel not within any cell)
    """

    if len(origin) == 3:
        origin = tuple(origin[0:2])
    assert len(origin) == 2
    assert width > 0 and height > 0
    if dy is None:
        dy = dx
    assert dx > 0.0 and dy > 0.0
    if k0 is not None:
        assert 0 <= k0 < grid.nk
    assert vertical_ref in ['top', 'base']

    kp = 0 if vertical_ref == 'top' else 1
    if k0 is not None:
        hp = grid.split_horizon_points(ref_k0 = k0, masked = False, kp = kp)
        p_map = grid.pixel_map_for_split_horizon_points(hp, origin, width, height, dx, dy = dy)
    else:
        _, _, raw_k = grid.extract_k_gaps()
        hp = grid.split_horizons_points(masked = False)
        p_map = np.empty((grid.nk, height, width, 2), dtype = int)
        for k0 in range(grid.nk):
            rk0 = raw_k[k0] + kp
            p_map[k0] = grid.pixel_map_for_split_horizon_points(hp[rk0], origin, width, height, dx, dy = dy)
    return p_map


def pixel_map_for_split_horizon_points(grid, horizon_points, origin, width, height, dx, dy = None):
    """Makes a mapping from pixels to cell j, i indices, based on split horizon points for a single horizon.

    args:
       horizon_points (numpy array of shape (nj, ni, 2, 2, 2+)): corner point x,y,z values for cell
          corners (j, i, jp, ip); as returned by split_horizon_points()
       origin (float pair): x, y of south west corner of area covered by pixel rectangle, in local crs
       width (int): the width of the pixel rectangle (number of pixels)
       height (int): the height of the pixel rectangle (number of pixels)
       dx (float): the size (west to east) of a pixel, in locel crs
       dx (float, optional): the size (south to north) of a pixel, in locel crs; defaults to dx

    returns:
       numpy int array of shape (height, width, 2), being the j, i indices of cells that the pixel centres lie within;
       values of -1 are used as null (ie. pixel not within any cell)
    """

    if dy is None:
        dy = dx
    half_dx = 0.5 * dx
    half_dy = 0.5 * dy
    d = np.array((dx, dy))
    half_d = np.array((half_dx, half_dy))

    #     north_east = np.array(origin) + np.array((width * dx, height * dy))

    p_map = np.full((height, width, 2), -1, dtype = int)

    # switch from logical corner ordering to polygon ordering
    poly_points = horizon_points[..., :2].copy()
    poly_points[:, :, 1, 1] = horizon_points[:, :, 1, 0, :2]
    poly_points[:, :, 1, 0] = horizon_points[:, :, 1, 1, :2]
    poly_points = poly_points.reshape(horizon_points.shape[0], horizon_points.shape[1], 4, 2)

    poly_box = np.empty((2, 2))
    patch_p_origin = np.empty((2,), dtype = int)  # NB. ordering is (ncol, nrow)
    patch_origin = np.empty((2,))
    patch_extent = np.empty((2,), dtype = int)  # NB. ordering is (ncol, nrow)

    for j in range(poly_points.shape[0]):
        for i in range(poly_points.shape[1]):
            if np.any(np.isnan(poly_points[j, i])):
                continue
            poly_box[0] = np.min(poly_points[j, i], axis = 0) - half_d
            poly_box[1] = np.max(poly_points[j, i], axis = 0) + half_d
            patch_p_origin[:] = (poly_box[0] - origin) / (dx, dy)
            if patch_p_origin[0] < 0 or patch_p_origin[1] < 0:
                continue
            patch_extent[:] = np.ceil((poly_box[1] - poly_box[0]) / (dx, dy))
            if patch_p_origin[0] + patch_extent[0] > width or patch_p_origin[1] + patch_extent[1] > height:
                continue
            patch_origin = origin + d * patch_p_origin + half_d
            scan_mask = pip.scan(patch_origin, patch_extent[0], patch_extent[1], dx, dy, poly_points[j, i])
            patch_mask = np.stack((scan_mask, scan_mask), axis = -1)
            old_patch = p_map[patch_p_origin[1]:patch_p_origin[1] + patch_extent[1],
                              patch_p_origin[0]:patch_p_origin[0] + patch_extent[0], :].copy()
            new_patch = np.empty(old_patch.shape, dtype = int)
            new_patch[:, :] = (j, i)
            p_map[patch_p_origin[1]:patch_p_origin[1] + patch_extent[1],
            patch_p_origin[0]:patch_p_origin[0] + patch_extent[0], :] = \
                np.where(patch_mask, new_patch, old_patch)
    return p_map



"""functions to handle different grid types"""

import resqpy
from resqpy.olio import xml_et as rqet


def grid_flavour(grid_root):
    """Returns a string indicating type of grid geometry, currently 'IjkGrid' or 'IjkBlockGrid'."""

    if grid_root is None:
        return None
    em = rqet.load_metadata_from_xml(grid_root)
    flavour = em.get('grid_flavour')
    if flavour is None:
        node_type = rqet.node_type(grid_root, strip_obj = True)
        if node_type == 'IjkGridRepresentation':
            if rqet.find_tag(grid_root, 'Geometry') is not None:
                flavour = 'IjkGrid'
            else:
                flavour = 'IjkBlockGrid'  # this might cause issues
        elif node_type == 'UnstructuredGridRepresentation':
            cell_shape = rqet.find_nested_tags_text(grid_root, ['Geometry', 'CellShape'])
            if cell_shape is None or cell_shape == 'polyhedral':
                flavour = 'UnstructuredGrid'
            elif cell_shape == 'tetrahedral':
                flavour = 'TetraGrid'
            elif cell_shape == 'hexahedral':
                flavour = 'HexaGrid'
            elif cell_shape == 'pyramidal':
                flavour = 'PyramidGrid'
            elif cell_shape == 'prism':
                flavour = 'PrismGrid'
    return flavour


def is_regular_grid(grid_root):
    """Returns True if the xml root node is for a RegularGrid."""

    return grid_flavour(grid_root) == 'IjkBlockGrid'


def any_grid(parent_model, grid_root = None, uuid = None, find_properties = True):
    """Returns a Grid or RegularGrid or UnstructuredGrid object depending on the extra metadata in the xml."""

    import resqpy.unstructured as rug

    if uuid is None and grid_root is not None:
        uuid = rqet.uuid_for_part_root(grid_root)
    flavour = grid_flavour(parent_model.root_for_uuid(uuid))
    if flavour is None:
        return None
    if flavour == 'IjkGrid':
        return resqpy.grid.Grid(parent_model, uuid = uuid, find_properties = find_properties)
    if flavour == 'IjkBlockGrid':
        return resqpy.grid.RegularGrid(parent_model, extent_kji = None, uuid = uuid, find_properties = find_properties)
    if flavour == 'UnstructuredGrid':
        return rug.UnstructuredGrid(parent_model, uuid = uuid, find_properties = find_properties)
    if flavour == 'TetraGrid':
        return rug.TetraGrid(parent_model, uuid = uuid, find_properties = find_properties)
    if flavour == 'HexaGrid':
        return rug.HexaGrid(parent_model, uuid = uuid, find_properties = find_properties)
    if flavour == 'PyramidGrid':
        return rug.PyramidGrid(parent_model, uuid = uuid, find_properties = find_properties)
    if flavour == 'PrismGrid':
        return rug.PrismGrid(parent_model, uuid = uuid, find_properties = find_properties)
    return None



"""Class for generic RESQML Organization Feature objects."""

from ._utils import alias_for_attribute, equivalent_extra_metadata
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class OrganizationFeature(BaseResqpy):
    """Class for generic RESQML Organization Feature objects."""

    resqml_type = "OrganizationFeature"
    feature_name = alias_for_attribute("title")

    def __init__(self,
                 parent_model,
                 uuid = None,
                 feature_name = None,
                 organization_kind = None,
                 originator = None,
                 extra_metadata = None):
        """Initialises an organization feature object."""

        self.organization_kind = organization_kind
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = feature_name,
                         originator = originator,
                         extra_metadata = extra_metadata)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this feature is essentially the same as the other; otherwise False."""

        if not isinstance(other, OrganizationFeature):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        return (self.feature_name == other.feature_name and self.organization_kind == other.organization_kind and
                ((not check_extra_metadata) or equivalent_extra_metadata(self, other)))

    def _load_from_xml(self):
        self.organization_kind = rqet.find_tag_text(self.root, 'OrganizationKind')

    def create_xml(self, add_as_part = True, originator = None, reuse = True):
        """Creates an organization feature xml node from this organization feature object."""

        if reuse and self.try_reuse():
            return self.root  # check for reusable (equivalent) object
        # create node with citation block
        ofn = super().create_xml(add_as_part = False, originator = originator)

        # Extra element for organization_kind
        if self.organization_kind not in ['earth model', 'fluid', 'stratigraphic', 'structural']:
            raise ValueError(self.organization_kind)
        kind_node = rqet.SubElement(ofn, ns['resqml2'] + 'OrganizationKind')
        kind_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'OrganizationKind')
        kind_node.text = self.organization_kind

        if add_as_part:
            self.model.add_part('obj_OrganizationFeature', self.uuid, ofn)

        return ofn



"""_wellbore_frame.py: resqpy well module providing wellbore frame class"""

import resqpy.well

# Nexus is a registered trademark of the Halliburton Company
# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.uuid as bu
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
import resqpy.property as rqp
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns

from .well_utils import load_hdf5_array


class WellboreFrame(BaseResqpy):
    """Class for RESQML WellboreFrameRepresentation objects (supporting well log Properties)

    RESQML documentation:

       Representation of a wellbore that is organized along a wellbore trajectory by its MD values.
       RESQML uses MD values to associate properties on points and to organize association of
       properties on intervals between MD points.

    Roughly equivalent to a Techlog "dataset" object with a given depth reference.

    The `logs` attribute is a :class:`resqpy.property.WellLogCollection` of all logs in the frame.
    """

    resqml_type = 'WellboreFrameRepresentation'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 trajectory = None,
                 mds = None,
                 represented_interp = None,
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Creates a new wellbore frame object and optionally loads it from xml or list of measured depths.

        arguments:
           parent_model (model.Model object): the model which the new wellbore frame belongs to
           uuid (optional): the uuid of an existing wellbore frame; if present, remaining arguments are
              ignored
           trajectory (Trajectory object, optional): the trajectory of the well; required if loading from
              list of measured depths
           mds (optional numpy 1D array, tuple or list of floats): ordered list of measured depths which
              will constitute the frame; ignored if uuid is not None
           represented_interp (wellbore interpretation object, optional): if present, is noted as the wellbore
              interpretation object which this frame relates to; ignored if uuid is not None
           title (str, optional): the citation title to use for a new wellbore frame;
              ignored if uuid is not None
           originator (str, optional): the name of the person creating the wellbore frame, defaults to login id;
              ignored if uuid is not None
           extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the wellbore frame;
              ignored if uuid is not None

        returns:
           the newly created wellbore frame object

        note:
           if initialising from a list of measured depths, the wellbore trajectory object must already exist
        """

        #: Associated wellbore trajectory, an instance of :class:`resqpy.well.Trajectory`.
        self.trajectory = trajectory
        self.trajectory_uuid = None if trajectory is None else trajectory.uuid

        #: Instance of :class:`resqpy.organize.WellboreInterpretation`
        self.wellbore_interpretation = represented_interp
        self.wellbore_feature = None
        self.feature_and_interpretation_to_be_written = False

        #: number of measured depth nodes, each being an entry or exit point of trajectory with a cell
        self.node_count = None

        #: node_count measured depths (in same units and datum as trajectory) of cell entry and/or exit points
        self.node_mds = None

        #: All logs associated with the wellbore frame; an instance of :class:`resqpy.property.WellLogCollection`
        self.logs = None

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is None and trajectory is not None and mds is not None and len(mds) > 1:
            self.node_count = len(mds)
            self.node_mds = np.array(mds)
            assert self.node_mds is not None and self.node_mds.ndim == 1

        # UUID needs to have been created before LogCollection can be made
        # TODO: Figure out when this should be created, and how it is kept in sync when new logs are created
        self.logs = rqp.WellLogCollection(frame = self)

    def _load_from_xml(self):
        """Loads the wellbore frame object from an xml node (and associated hdf5 data)."""

        # NB: node is the root level xml node, not a node in the md list!

        node = self.root
        assert node is not None

        trajectory_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(node, ['Trajectory', 'UUID']))
        assert trajectory_uuid is not None, 'wellbore frame trajectory reference not found in xml'
        if self.trajectory is None:
            self.trajectory = resqpy.well.Trajectory(self.model, uuid = trajectory_uuid)
        else:
            assert bu.matching_uuids(self.trajectory.uuid, trajectory_uuid), 'wellbore frame trajectory uuid mismatch'

        self.node_count = rqet.find_tag_int(node, 'NodeCount')
        assert self.node_count is not None, 'node count not found in xml for wellbore frame'
        assert self.node_count > 1, 'fewer than 2 nodes for wellbore frame'

        mds_node = rqet.find_tag(node, 'NodeMd')
        assert mds_node is not None, 'wellbore frame measured depths hdf5 reference not found in xml'
        load_hdf5_array(self, mds_node, 'node_mds')

        assert self.node_mds is not None and self.node_mds.ndim == 1 and self.node_mds.size == self.node_count

        interp_uuid = rqet.find_nested_tags_text(node, ['RepresentedInterpretation', 'UUID'])
        if interp_uuid is None:
            self.wellbore_interpretation = None
        else:
            self.wellbore_interpretation = rqo.WellboreInterpretation(self.model, uuid = interp_uuid)

        # Create well log collection of all log data
        self.logs = rqp.WellLogCollection(frame = self)
        # pass

    def extract_crs_uuid(self):
        """Returns the uuid of the coordinate reference system used by the related trajectory."""

        if self.trajectory is None:
            return None
        return self.trajectory.crs_uuid

    def create_feature_and_interpretation(self):
        """Instantiate new empty WellboreFeature and WellboreInterpretation objects, if a wellboreinterpretation does

        not already exist.

        Uses the wellboreframe citation title as the well name
        """

        if self.wellbore_interpretation is None:
            log.info(f"Creating WellboreInterpretation and WellboreFeature with name {self.title}")
            self.wellbore_feature = rqo.WellboreFeature(parent_model = self.model, feature_name = self.title)
            self.wellbore_interpretation = rqo.WellboreInterpretation(parent_model = self.model,
                                                                      wellbore_feature = self.wellbore_feature)
            self.feature_and_interpretation_to_be_written = True
        else:
            log.info("WellboreInterpretation already exists")

    def write_hdf5(self, file_name = None, mode = 'a'):
        """Create or append to an hdf5 file, writing datasets for the measured depths."""

        # NB: array data must have been set up prior to calling this function

        if self.uuid is None:
            self.uuid = bu.new_uuid()

        h5_reg = rwh5.H5Register(self.model)
        h5_reg.register_dataset(self.uuid, 'NodeMd', self.node_mds)
        h5_reg.write(file = file_name, mode = mode)

    def create_xml(self,
                   ext_uuid = None,
                   add_as_part = True,
                   add_relationships = True,
                   title = None,
                   originator = None):
        """Create a wellbore frame representation node from this WellboreFrame object, optionally add as part.

        note:
           trajectory xml node must be in place before calling this function
        """

        assert self.trajectory is not None, 'trajectory object missing'
        assert self.trajectory.root is not None, 'trajectory xml not established'

        self.__create_wellbore_feature_and_interpretation_xml(add_as_part = add_as_part,
                                                              add_relationships = add_relationships,
                                                              originator = originator)

        if ext_uuid is None:
            ext_uuid = self.model.h5_uuid()

        if title:
            self.title = title
        if not self.title:
            self.title = 'wellbore frame'

        wf_node = super().create_xml(originator = originator, add_as_part = False)

        # wellbore frame elements

        nc_node, mds_node, mds_values_node = self.__add_sub_elements_to_root_node(wf_node = wf_node)

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'NodeMd', root = mds_values_node)

        traj_root = self.trajectory.root
        self.model.create_ref_node('Trajectory',
                                   rqet.find_nested_tags_text(traj_root, ['Citation', 'Title']),
                                   bu.uuid_from_string(traj_root.attrib['uuid']),
                                   content_type = 'obj_WellboreTrajectoryRepresentation',
                                   root = wf_node)

        if self.wellbore_interpretation is not None:
            interp_root = self.wellbore_interpretation.root
            self.model.create_ref_node('RepresentedInterpretation',
                                       rqet.find_nested_tags_text(interp_root, ['Citation', 'Title']),
                                       bu.uuid_from_string(interp_root.attrib['uuid']),
                                       content_type = 'obj_WellboreInterpretation',
                                       root = wf_node)

        self.__add_as_part_and_add_relationships(wf_node = wf_node,
                                                 ext_uuid = ext_uuid,
                                                 add_as_part = add_as_part,
                                                 add_relationships = add_relationships)

        return wf_node

    def __create_wellbore_feature_and_interpretation_xml(self,
                                                         add_as_part = True,
                                                         add_relationships = True,
                                                         originator = None):
        """ Return root node for WellboreFeature and WellboreInterpretation objects"""

        if self.feature_and_interpretation_to_be_written:
            if self.wellbore_interpretation is None:
                self.create_feature_and_interpretation()
            if self.wellbore_feature is not None:
                self.wellbore_feature.create_xml(add_as_part = add_as_part, originator = originator)
            self.wellbore_interpretation.create_xml(add_as_part = add_as_part,
                                                    add_relationships = add_relationships,
                                                    originator = originator)

    def __add_sub_elements_to_root_node(self, wf_node):
        """Appends sub-elements to the WellboreFrame object's root node."""

        nc_node = rqet.SubElement(wf_node, ns['resqml2'] + 'NodeCount')
        nc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        nc_node.text = str(self.node_count)

        mds_node = rqet.SubElement(wf_node, ns['resqml2'] + 'NodeMd')
        mds_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'DoubleHdf5Array')
        mds_node.text = rqet.null_xml_text

        mds_values_node = rqet.SubElement(mds_node, ns['resqml2'] + 'Values')
        mds_values_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Hdf5Dataset')
        mds_values_node.text = rqet.null_xml_text

        return nc_node, mds_node, mds_values_node

    def __add_as_part_and_add_relationships(self, wf_node, ext_uuid, add_as_part, add_relationships):
        """Add the newly created WellborFrame object's root node as a part in the model and add reciprocal relationships.."""

        if add_as_part:
            self.model.add_part('obj_WellboreFrameRepresentation', self.uuid, wf_node)
            if add_relationships:
                self.model.create_reciprocal_relationship(wf_node, 'destinationObject', self.trajectory.root,
                                                          'sourceObject')
                ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                ext_node = self.model.root_for_part(ext_part)
                self.model.create_reciprocal_relationship(wf_node, 'mlToExternalPartProxy', ext_node,
                                                          'externalPartProxyToMl')
                if self.wellbore_interpretation is not None:
                    interp_root = self.wellbore_interpretation.root
                    self.model.create_reciprocal_relationship(wf_node, 'destinationObject', interp_root, 'sourceObject')



"""Multiprocessing wrapper functions for the grid_surface module."""

import logging

log = logging.getLogger(__name__)

import numpy as np
from typing import Tuple, Union, List, Optional, Callable
import resqpy.grid_surface as rqgs
from resqpy.model import new_model
from resqpy.grid import RegularGrid
from resqpy.surface import Surface, PointSet
from resqpy.property import PropertyCollection
from pathlib import Path
from resqpy.model import Model
import resqpy.olio.uuid as bu
from uuid import UUID
import uuid


def find_faces_to_represent_surface_regular_wrapper(
    index: int,
    use_index_as_realisation: bool,
    grid_epc: str,
    grid_uuid: Union[UUID, str],
    surface_epc: str,
    surface_uuid: Union[UUID, str],
    name: str,
    title: Optional[str] = None,
    agitate: bool = False,
    feature_type: str = 'fault',
    trimmed: bool = False,
    extend_fault_representation: bool = False,
    related_uuid = None,
    progress_fn: Optional[Callable] = None,
    consistent_side: bool = False,
    extra_metadata = None,
    return_properties: Optional[List[str]] = None,
) -> Tuple[int, bool, str, List[Union[UUID, str]]]:
    """Wrapper function of find_faces_to_represent_surface_regular_optimised.

    Used for multiprocessing to create a new model that is saved in a temporary epc file
    and returns the required values, which are used in the multiprocessing function to
    recombine all the objects into a single epc file.

    Args:
        index (int): the index of the function call from the multiprocessing function.
        use_index_as_realisation (bool): if True, uses the index number as the realization number on
            the property collection.
        grid_epc (str): epc file path where the grid is saved.
        grid_uuid (UUID/str): UUID (universally unique identifier) of the grid object.
        surface_epc (str): epc file path where the surface (or point set) is saved.
        surface_uuid (UUID/str): UUID (universally unique identifier) of the surface (or point set) object.
        name (str): the feature name to use in the grid connection set.
        title (str): the citation title to use for the grid connection set; defaults to name
        agitate (bool): if True, the points of the surface are perturbed by a small random
           offset, which can help if the surface has been built from a regular mesh with a periodic resonance
           with the grid
        feature_type (str, default 'fault'): one of 'fault', 'horizon', or 'geobody boundary'
        trimmed (bool, default True): if True the surface has already been trimmed
        extend_fault_representation (bool, default False): if True, the representation is extended with a flange
        related_uuid (uuid, optional): if present, the uuid of an object to be softly related to the gcs
        progress_fn (Callable): a callback function to be called at intervals by this function;
           the argument will progress from 0.0 to 1.0 in unspecified and uneven increments
        consistent_side (bool): if True, the cell pairs will be ordered so that all the first
           cells in each pair are on one side of the surface, and all the second cells on the other
        extra_metadata (dict, optional): extra metadata items to be added to the grid connection set
        return_properties (List[str]): if present, a list of property arrays to calculate and
           return as a dictionary; recognised values in the list are 'triangle', 'depth', 'offset' and 'normal vector';
           triangle is an index into the surface triangles of the triangle detected for the gcs face; depth is
           the z value of the intersection point of the inter-cell centre vector with a triangle in the surface;
           offset is a measure of the distance between the centre of the cell face and the intersection point;
           normal vector is a unit vector normal to the surface triangle; each array has an entry for each face
           in the gcs; the returned dictionary has the passed strings as keys and numpy arrays as values.

    Returns:
        Tuple containing:

            - index (int): the index passed to the function.
            - success (bool): whether the function call was successful, whatever that
                definiton is.
            - epc_file (str): the epc file path where the objects are stored.
            - uuid_list (List[str]): list of UUIDs of relevant objects.
    """
    tmp_dir = Path(f"tmp_dir/{uuid.uuid4()}")
    tmp_dir.mkdir(parents = True, exist_ok = True)
    epc_file = f"{tmp_dir}/wrapper.epc"
    model = new_model(epc_file = epc_file)
    uuid_list = []
    g_model = Model(grid_epc)
    g_crs_uuid = g_model.uuid(obj_type = 'LocalDepth3dCrs',
                              related_uuid = grid_uuid)  # todo: check this relationship exists
    if g_crs_uuid is not None:
        model.copy_uuid_from_other_model(g_model, g_crs_uuid)
        uuid_list.append(g_crs_uuid)
    model.copy_uuid_from_other_model(g_model, uuid = grid_uuid)
    uuid_list.append(grid_uuid)
    for prop_title in ['DX', 'DY', 'DZ']:
        prop_uuid = g_model.uuid(obj_type = 'ContinuousProperty', title = prop_title, related_uuid = grid_uuid)
        if prop_uuid is not None:
            log.debug(f'found grid cell length property {prop_title}')
            model.copy_uuid_from_other_model(g_model, uuid = prop_uuid)
            uuid_list.append(prop_uuid)
        else:
            log.warning(f'grid cell length property {prop_title} NOT found')
    grid = RegularGrid(parent_model = model, uuid = grid_uuid)
    s_model = Model(surface_epc)
    model.copy_uuid_from_other_model(s_model, uuid = str(surface_uuid))
    repr_type = model.type_of_part(model.part(uuid = surface_uuid), strip_obj = True)
    assert repr_type in ['TriangulatedSetRepresentation', 'PointSetRepresentation']
    extended = False
    if repr_type == 'PointSetRepresentation':
        # trim pointset to grid xyz box
        pset = PointSet(model, uuid = surface_uuid)
        log.debug(f'point set {pset.title} raw point count: {len(pset.full_array_ref())}')
        pset.change_crs(grid.crs)
        if not trimmed:
            pset.trim_to_xyz_box(grid.xyz_box(local = True))
            trimmed = True
        assert len(pset.full_array_ref()) >= 3,  \
            f'boundary {name} representation {pset.title} has no xyz overlap with grid'
        pset_points = pset.full_array_ref()
        log.debug(f'trimmed point set {pset.title} contains {len(pset_points)} points; about to triangulate')
        if len(pset_points) > 1000:
            log.warning(
                f'trimmed point set {pset.title} has {len(pset_points)} points, which might take a while to triangulate'
            )
        # triangulate point set to form a surface; set repr_uuid to that surface and switch repr_flavour to 'surface'
        surf = Surface(model, crs_uuid = grid.crs.uuid, title = pset.title)
        surf.set_from_point_set(pset,
                                convexity_parameter = 2.0,
                                reorient = True,
                                extend_with_flange = extend_fault_representation,
                                make_clockwise = False)
        extended = extend_fault_representation
        surf.write_hdf5()
        surf.create_xml()
        surface_uuid = surf.uuid
    surface = Surface(parent_model = model, uuid = str(surface_uuid))
    surface.change_crs(grid.crs)
    if not trimmed and surface.triangle_count() > 100:
        trimmed_surf = Surface(model, crs_uuid = grid.crs.uuid)
        # trimmed_surf.set_to_trimmed_surface(surf, xyz_box = xyz_box, xy_polygon = parent_seg.polygon)
        trimmed_surf.set_to_trimmed_surface(surface, xyz_box = grid.xyz_box(local = True))
        surface = trimmed_surf
        trimmed = True
    if extend_fault_representation and not extended:
        _, p = surface.triangles_and_points()
        pset = PointSet(model, points_array = p, crs_uuid = grid.crs.uuid, title = surface.title)
        surface = Surface(model, crs_uuid = grid.crs.uuid, title = pset.title)
        surface.set_from_point_set(pset,
                                   convexity_parameter = 2.0,
                                   reorient = True,
                                   extend_with_flange = True,
                                   make_clockwise = False)
        extended = True
    if not bu.matching_uuids(surface.uuid, surface_uuid):
        surface.write_hdf5()
        surface.create_xml()
        surface_uuid = surface.uuid
    uuid_list.append(surface_uuid)

    returns = rqgs.find_faces_to_represent_surface_regular_optimised(
        grid,
        surface,
        name,
        title,
        None,  # centres
        agitate,
        feature_type,
        progress_fn,
        consistent_side,
        return_properties,
    )

    success = False

    if isinstance(returns, tuple):
        gcs = returns[0]
    else:
        gcs = returns

    if gcs.count > 0:
        success = True
        gcs.write_hdf5()
        gcs.create_xml(extra_metadata = extra_metadata)
        model.copy_uuid_from_other_model(gcs.model, uuid = gcs.uuid)
        if related_uuid is not None:
            model.create_reciprocal_relationship_uuids(gcs.uuid, 'sourceObject', related_uuid, 'destinationObject')
        uuid_list.append(gcs.uuid)

    if success and return_properties is not None and len(return_properties):
        log.debug(f'{name} requested properties: {return_properties}')
        properties = returns[1]
        realisation = index if use_index_as_realisation else None
        property_collection = PropertyCollection(support = gcs)
        for p_name, array in properties.items():
            log.debug(f'{name} found property {p_name}')
            if p_name == "normal vector":
                property_collection.add_cached_array_to_imported_list(
                    array,
                    "from find_faces function",
                    p_name,
                    discrete = False,
                    uom = "Euc",
                    property_kind = "continuous",
                    realization = realisation,
                    indexable_element = "faces",
                    points = True,
                )
            elif p_name == "triangle":
                property_collection.add_cached_array_to_imported_list(
                    array,
                    "from find_faces function",
                    p_name,
                    discrete = True,
                    null_value = -1,
                    property_kind = "discrete",
                    realization = realisation,
                    indexable_element = "faces",
                )
            elif p_name == "offset":
                property_collection.add_cached_array_to_imported_list(
                    array,
                    "from find_faces function",
                    p_name,
                    discrete = False,
                    uom = grid.crs.z_units,
                    property_kind = "continuous",
                    realization = realisation,
                    indexable_element = "faces",
                )
            elif p_name == "depth":
                # convert values to global z inc down
                array[:] += grid.crs.z_offset
                if not grid.crs.z_inc_down:
                    array = -array
                property_collection.add_cached_array_to_imported_list(
                    array,
                    "from find_faces function",
                    p_name,
                    discrete = False,
                    uom = grid.crs.z_units,
                    property_kind = "depth",
                    realization = realisation,
                    indexable_element = "faces",
                )
        property_collection.write_hdf5_for_imported_list()
        uuids_properties = (property_collection.create_xml_for_imported_list_and_add_parts_to_model())
        uuid_list.extend(uuids_properties)
    else:
        log.debug(f'{name} no requested properties')

    model.store_epc()

    return index, success, epc_file, uuid_list



"""Containing resqml property class"""

version = '27th June 2022'

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy

from .property_collection import PropertyCollection


class Property(BaseResqpy):
    """Class for an individual property object; uses a single element PropertyCollection behind the scenes."""

    @property
    def resqml_type(self):
        """Returns the RESQML object class string for this Property."""
        root_node = self.root
        if root_node is not None:
            return rqet.node_type(root_node, strip_obj = True)
        if (not hasattr(self, 'collection') or self.collection.number_of_parts() != 1 or self.is_continuous()):
            return 'ContinuousProperty'
        if self.collection.points_for_part(self.collection.parts()[0]):
            return 'PointsProperty'
        return 'CategoricalProperty' if self.is_categorical() else 'DiscreteProperty'

    def __init__(self, parent_model, uuid = None, title = None, support_uuid = None, extra_metadata = None):
        """Initialises a resqpy Property object, either for an existing RESQML property, or empty for support.

        arguments:
           parent_model (model.Model): the model to which the property belongs
           uuid (uuid.UUID, optional): required if initialising from an existing RESQML property object
           title (str, optional): the citation title to use for the property; ignored if uuid is present
           support_uuid (uuid.UUID, optional): identifies the supporting representation for the property;
              ignored if uuid is present
           extra_metadata (dict, optional): if present, the dictionary items are added as extra metadata;
              ignored if uuid is present

        returns:
           new resqpy Property object
        """

        self.collection = PropertyCollection()
        super().__init__(model = parent_model, uuid = uuid, title = title, extra_metadata = extra_metadata)
        if support_uuid is not None:
            if self.collection.support_uuid is None:
                self.collection.set_support(model = parent_model, support_uuid = support_uuid)
            else:
                assert bu.matching_uuids(support_uuid, self.collection.support_uuid)

    def _load_from_xml(self):
        """Populates this property object from xml; not usually called directly."""

        part = self.part
        assert part is not None
        if self.collection is None:
            self.collection = PropertyCollection()
        if self.collection.model is None:
            self.collection.model = self.model
        self.collection.add_part_to_dict(part)
        # duplicate extra metadata, as standard attribute in BaseResqpy
        self.extra_metadata = self.collection.extra_metadata_for_part(part)
        self.collection.has_single_property_kind_flag = True
        self.collection.has_single_indexable_element_flag = True
        self.collection.has_multiple_realizations_flag = False
        assert self.collection.number_of_parts() == 1

    @classmethod
    def from_singleton_collection(cls, property_collection: PropertyCollection):
        """Populates a new Property from a PropertyCollection containing just one part.

        arguments:
           property_collection (PropertyCollection): the singleton collection from which to populate this Property
        """

        # Validate
        assert property_collection.model is not None
        assert property_collection is not None
        assert property_collection.number_of_parts() == 1

        # Instantiate the object i.e. call the class __init__ method
        part = property_collection.parts()[0]
        prop = cls(parent_model = property_collection.model,
                   uuid = property_collection.uuid_for_part(part),
                   support_uuid = property_collection.support_uuid,
                   title = property_collection.citation_title_for_part(part),
                   extra_metadata = property_collection.extra_metadata_for_part(part))

        # Edit properties of collection attribute
        prop.collection.has_single_property_kind_flag = True
        prop.collection.has_single_indexable_element_flag = True
        prop.collection.has_multiple_realizations_flag = False

        return prop

    @classmethod
    def from_array(cls,
                   parent_model,
                   cached_array,
                   source_info,
                   keyword,
                   support_uuid,
                   property_kind = None,
                   local_property_kind_uuid = None,
                   indexable_element = None,
                   facet_type = None,
                   facet = None,
                   discrete = False,
                   uom = None,
                   null_value = None,
                   time_series_uuid = None,
                   time_index = None,
                   realization = None,
                   count = 1,
                   points = False,
                   const_value = None,
                   string_lookup_uuid = None,
                   find_local_property_kind = True,
                   expand_const_arrays = False,
                   dtype = None,
                   extra_metadata = {}):
        """Populates a new Property from a numpy array and metadata; NB. Writes data to hdf5 and adds part to model.

        arguments:
           parent_model (model.Model): the model to which the new property belongs
           cached_array: a numpy array to be made into a property; for a constant array set cached_array to None
              (and use const_value)
           source_info (string): typically the name of a file from which the array has been read but can be any
              information regarding the source of the data
           keyword (string): this will be used as the citation title for the property
           support_uuid (uuid): the uuid of the supporting representation
           property_kind (string): resqml property kind (required unless a local propery kind is identified with
              local_property_kind_uuid)
           local_property_kind_uuid (uuid.UUID or string, optional): uuid of local property kind, or None if a
              standard property kind applies or if the local property kind is identified by its name (see
              find_local_property_kind)
           indexable_element (string): the indexable element in the supporting representation; if None
              then the 'typical' indexable element for the class of supporting representation will be assumed
           facet_type (string, optional): resqml facet type, or None; resqpy only supports at most one facet per
              property though RESQML allows for multiple)
           facet (string, optional): resqml facet, or None; resqpy only supports at most one facet per property
              though RESQML allows for multiple)
           discrete (boolean, optional, default False): if True, the array should contain integer (or boolean)
              data, if False, float; set True for any discrete data including categorical
           uom (string, optional, default None): the resqml units of measure for the data; required for
              continuous (float) array
           null_value (int, optional, default None): if present, this is used in the metadata to indicate that
              the value is to be interpreted as a null value wherever it appears in the discrete data; not used
              for continuous data where NaN is always the null value
           time_series_uuid (optional): the uuid of the full or reduced time series that the time_index is for
           time_index (integer, optional, default None): if not None, the time index for the property (see also
              time_series_uuid)
           realization (int, optional): realization number, or None
           count (int, default 1): the number of values per indexable element; if greater than one then this
              must be the fastest cycling axis in the cached array, ie last index
           const_value (int, float or bool, optional): the value with which a constant array is filled;
              required if cached_array is None, must be None otherwise
           string_lookup_uuid (optional): if present, the uuid of the string table lookup which the categorical data
              relates to; if None, the property will not be configured as categorical
           find_local_property_kind (boolean, default True): if True, local property kind uuid need not be provided as
              long as the property_kind is set to match the title of the appropriate local property kind object
           expand_const_arrays (boolean, default False): if True, and a const_value is given, the array will be fully
              expanded and written to the hdf5 file; the xml will then not indicate that it is constant
           dtype (numpy dtype, optional): if present, the elemental data type to use when writing the array to hdf5
           extra_metadata (optional): if present, a dictionary of extra metadata to be added for the part

        returns:
           new Property object built from numpy array; the hdf5 data has been written, xml created and the part
           added to the model

        notes:
           this method writes to the hdf5 file and creates the xml node, which is added as a part to the model;
           calling code must still call the model's store_epc() method;
           this from_array() method is a convenience method calling self.collection.set_support(),
           self.prepare_import(), self.write_hdf5() and self.create_xml()

        :meta common:
        """
        # Validate
        assert parent_model is not None
        assert cached_array is not None or const_value is not None

        # Instantiate the object i.e. call the class __init__ method
        prop = cls(parent_model = parent_model,
                   title = keyword,
                   support_uuid = support_uuid,
                   extra_metadata = extra_metadata)

        # Prepare array data in collection attribute and add to model
        prop.collection.set_support(model = prop.model,
                                    support_uuid = support_uuid)  # this can be expensive; todo: optimise
        if prop.collection.model is None:
            prop.collection.model = prop.model
        prop.prepare_import(cached_array,
                            source_info,
                            keyword,
                            discrete = discrete,
                            uom = uom,
                            time_index = time_index,
                            null_value = null_value,
                            property_kind = property_kind,
                            local_property_kind_uuid = local_property_kind_uuid,
                            facet_type = facet_type,
                            facet = facet,
                            realization = realization,
                            indexable_element = indexable_element,
                            count = count,
                            points = points,
                            const_value = const_value)
        prop.write_hdf5(expand_const_arrays = expand_const_arrays, dtype = dtype)
        prop.create_xml(support_uuid = support_uuid,
                        time_series_uuid = time_series_uuid,
                        string_lookup_uuid = string_lookup_uuid,
                        property_kind_uuid = local_property_kind_uuid,
                        find_local_property_kind = find_local_property_kind,
                        expand_const_arrays = expand_const_arrays,
                        extra_metadata = extra_metadata)
        return prop

    def array_ref(self, dtype = None, masked = False, exclude_null = False):
        """Returns a (cached) numpy array containing the property values.

        arguments:
           dtype (str or type, optional): the required dtype of the array (usually float, int or bool)
           masked (boolean, default False): if True, a numpy masked array is returned, with the mask determined by
              the inactive cell mask in the case of a Grid property
           exclude_null (boolean, default False): if True and masked is True, elements whose value is the null value
              (NaN for floats) will be masked out

        returns:
           numpy array

        :meta common:
        """

        return self.collection.cached_part_array_ref(self.part,
                                                     dtype = dtype,
                                                     masked = masked,
                                                     exclude_null = exclude_null)

    def is_continuous(self):
        """Returns boolean indicating that the property contains continuous (ie. float) data.

        :meta common:
        """
        return self.collection.continuous_for_part(self.part)

    def is_points(self):
        """Returns boolean indicating that the property is a points property."""

        return self.collection.points_for_part(self.part)

    def is_categorical(self):
        """Returns boolean indicating that the property contains categorical data.

        :meta common:
        """
        return self.collection.part_is_categorical(self.part)

    def null_value(self):
        """Returns int being the null value for the (discrete) property."""
        return self.collection.null_value_for_part(self.part)

    def count(self):
        """Returns int being the number of values for each element (usually 1)."""
        return self.collection.count_for_part(self.part)

    def indexable_element(self):
        """Returns string being the indexable element for the property.

        :meta common:
        """
        return self.collection.indexable_for_part(self.part)

    def property_kind(self):
        """Returns string being the property kind for the property.

        :meta common:
        """
        return self.collection.property_kind_for_part(self.part)

    def local_property_kind_uuid(self):
        """Returns uuid of the local property kind for the property (if applicable, otherwise None)."""
        return self.collection.local_property_kind_uuid(self.part)

    def facet_type(self):
        """Returns string being the facet type for the property, or None if no facet present.

        note: resqpy currently supports at most one facet per property, though RESQML allows for multiple.

        :meta common:
        """
        return self.collection.facet_type_for_part(self.part)

    def facet(self):
        """Returns string being the facet value for the property, or None if no facet present.

        note: resqpy currently supports at most one facet per property, though RESQML allows for multiple.

        :meta common:
        """
        return self.collection.facet_for_part(self.part)

    def time_series_uuid(self):
        """Returns the uuid of the related time series (if applicable, otherwise None)."""
        return self.collection.time_series_uuid_for_part(self.part)

    def time_index(self):
        """Returns int being the index into the related time series (if applicable, otherwise None)."""
        return self.collection.time_index_for_part(self.part)

    def minimum_value(self):
        """Returns the minimum value for the property as stored in xml (float or int or None)."""
        return self.collection.minimum_value_for_part(self.part)

    def maximum_value(self):
        """Returns the maximum value for the property as stored in xml (float or int or None)."""
        return self.collection.maximum_value_for_part(self.part)

    def uom(self):
        """Returns string being the units of measure (for a continuous property, otherwise None).

        :meta common:
        """
        return self.collection.uom_for_part(self.part)

    def string_lookup_uuid(self):
        """Returns the uuid of the related string lookup table (for a categorical property, otherwise None)."""
        return self.collection.string_lookup_uuid_for_part(self.part)

    def string_lookup(self):
        """Returns a resqpy StringLookup table (for a categorical property, otherwise None)."""
        return self.collection.string_lookup_for_part(self.part)

    def constant_value(self):
        """For a constant property, returns the constant value (float or int or bool, or None if not constant)."""
        return self.collection.constant_value_for_part(self.part)

    def prepare_import(self,
                       cached_array,
                       source_info,
                       keyword,
                       discrete = False,
                       uom = None,
                       time_index = None,
                       null_value = None,
                       property_kind = None,
                       local_property_kind_uuid = None,
                       facet_type = None,
                       facet = None,
                       realization = None,
                       indexable_element = None,
                       count = 1,
                       const_value = None,
                       points = False):
        """Takes a numpy array and metadata and sets up a single array import list; not usually called directly.

        note:
           see the documentation for the convenience method from_array()
        """
        assert self.collection.number_of_parts() == 0
        assert not self.collection.imported_list
        self.collection.add_cached_array_to_imported_list(cached_array,
                                                          source_info,
                                                          keyword,
                                                          discrete = discrete,
                                                          uom = uom,
                                                          time_index = time_index,
                                                          null_value = null_value,
                                                          property_kind = property_kind,
                                                          local_property_kind_uuid = local_property_kind_uuid,
                                                          facet_type = facet_type,
                                                          facet = facet,
                                                          realization = realization,
                                                          indexable_element = indexable_element,
                                                          count = count,
                                                          const_value = const_value,
                                                          points = points)

    def write_hdf5(self, file_name = None, mode = 'a', expand_const_arrays = False, dtype = None):
        """Writes the array data to the hdf5 file; not usually called directly.

           file_name (str, optional): if present, the path of the hdf5 file to use; strongly recommended not to
              set this argument
           mode (str, default 'a'): 'a' or 'w' being the mode to open the hdf5 file in; strongly recommended to use 'a'
           expand_const_arrays (bool, default False): if True and the array is a constant array then a fully populated
              array is generated and stored (otherwise the constant value is held in xml and no hdf5 data is needed)
           dtype (numpy dtype, optional): if present, the elemental data type to use when writing the array to hdf5

        notes:
           see the documentation for the convenience method from_array();
        """
        if not self.collection.imported_list:
            log.warning('no imported Property array to write to hdf5')
            return
        self.collection.write_hdf5_for_imported_list(file_name = file_name,
                                                     mode = mode,
                                                     expand_const_arrays = expand_const_arrays,
                                                     dtype = dtype)

    def create_xml(self,
                   ext_uuid = None,
                   support_uuid = None,
                   time_series_uuid = None,
                   string_lookup_uuid = None,
                   property_kind_uuid = None,
                   find_local_property_kind = True,
                   expand_const_arrays = False,
                   extra_metadata = {}):
        """Creates an xml tree for the property and adds it as a part to the model; not usually called directly.

        note:
           see the documentation for the convenience method from_array()
           NB. this method has the deliberate side effect of modifying the uuid and title of self to match those
           of the property collection part!
        """
        if not self.collection.imported_list:
            log.warning('no imported Property array to create xml for')
            return
        self.collection.create_xml_for_imported_list_and_add_parts_to_model(
            ext_uuid = ext_uuid,
            support_uuid = support_uuid,
            time_series_uuid = time_series_uuid,
            selected_time_indices_list = None,
            string_lookup_uuid = string_lookup_uuid,
            property_kind_uuid = property_kind_uuid,
            find_local_property_kinds = find_local_property_kind,
            expand_const_arrays = expand_const_arrays,
            extra_metadata = extra_metadata)
        self.collection.has_single_property_kind_flag = True
        self.collection.has_single_uom_flag = True
        self.collection.has_single_indexable_element_flag = True
        self.collection.has_multiple_realizations_flag = False
        part = self.collection.parts()[0]
        assert part is not None
        self.uuid = self.collection.uuid_for_part(part)
        self.title = self.collection.citation_title_for_part(part)
        return self.root



"""RESQML Feature and Interpretation classes."""

version = '9th November 2021'

# most feature and interpretation classes catered for here
# stratigraphic classes in strata.py

import logging

log = logging.getLogger(__name__)
log.debug('organize_old.py version ' + version)

from ._utils import alias_for_attribute, equivalent_extra_metadata
import resqpy.olio.uuid as bu
from resqpy.olio.base import BaseResqpy


class GeobodyFeature(BaseResqpy):
    """Class for RESQML Geobody Feature objects (note: definition may be incomplete in RESQML 2.0.1)."""

    resqml_type = "GeobodyFeature"
    feature_name = alias_for_attribute("title")

    def __init__(self, parent_model, uuid = None, feature_name = None, extra_metadata = None):
        """Initialises a geobody feature object."""

        super().__init__(model = parent_model, uuid = uuid, title = feature_name, extra_metadata = extra_metadata)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this feature is essentially the same as the other; otherwise False."""

        if other is None or not isinstance(other, self.__class__):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if check_extra_metadata and not equivalent_extra_metadata(self, other):
            return False
        return self.feature_name == other.feature_name

    def create_xml(self, add_as_part = True, originator = None, reuse = True):
        """Creates a geobody feature xml node from this geobody feature object."""
        if reuse and self.try_reuse():
            return self.root  # check for reusable (equivalent) object
        return super().create_xml(add_as_part = add_as_part, originator = originator)



"""High level gather ensemble function."""

import logging

log = logging.getLogger(__name__)

from time import time

import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp


def gather_ensemble(case_epc_list,
                    new_epc_file,
                    consolidate = True,
                    shared_grids = True,
                    shared_time_series = True,
                    create_epc_lookup = True):
    """Creates a composite resqml dataset by merging all parts from all models in list, assigning realization numbers.

    arguments:
       case_epc_list (list of strings): paths of individual realization epc files
       new_epc_file (string): path of new composite epc to be created (with paired hdf5 file)
       consolidate (boolean, default True): if True, simple parts are tested for equivalence and where similar enough
          a single shared object is established in the composite dataset
       shared_grids (boolean, default True): if True and consolidate is True, then grids are also consolidated
          with equivalence based on extent of grids (and citation titles if grid extents within the first case
          are not distinct); ignored if consolidate is False
       shared_time_series (boolean, default False): if True and consolidate is True, then time series are consolidated
          with equivalence based on title, without checking that timestamp lists are the same
       create_epc_lookup (boolean, default True): if True, a StringLookupTable is created to map from realization
          number to case epc path

    notes:
       property objects will have an integer realization number assigned, which matches the corresponding index into
       the case_epc_list;
       if consolidating with shared grids, then only properties will be gathered from realisations after the first and
       an exception will be raised if the grids are not matched between realisations
    """

    if not consolidate:
        shared_grids = False

    composite_model = rq.Model(new_epc_file, new_epc = True, create_basics = True, create_hdf5_ext = True)

    epc_lookup_dict = {}

    for r, case_epc in enumerate(case_epc_list):
        t_r_start = time()  # debug
        log.info(f'gathering realszation {r}: {case_epc}')
        epc_lookup_dict[r] = case_epc
        case_model = rq.Model(case_epc)
        if r == 0:  # first case
            log.info('first case')  # debug
            composite_model.copy_all_parts_from_other_model(case_model, realization = 0, consolidate = consolidate)
            if shared_time_series:
                host_ts_uuids = case_model.uuids(obj_type = 'TimeSeries')
                host_ts_titles = []
                for ts_uuid in host_ts_uuids:
                    host_ts_titles.append(case_model.title(uuid = ts_uuid))
            if shared_grids:
                host_grid_uuids = case_model.uuids(obj_type = 'IjkGridRepresentation')
                host_grid_shapes = []
                host_grid_titles = []
                title_match_required = False
                for grid_uuid in host_grid_uuids:
                    grid_root = case_model.root(uuid = grid_uuid)
                    host_grid_shapes.append(grr.extent_kji_from_root(grid_root))
                    host_grid_titles.append(rqet.citation_title_for_node(grid_root))
                if len(set(host_grid_shapes)) < len(host_grid_shapes):
                    log.warning(
                        'shapes of representative grids are not distinct, grid titles must match during ensemble gathering'
                    )
                    title_match_required = True
        else:  # subsequent cases
            log.info('subsequent case')  # debug
            composite_model.consolidation = None  # discard any previous mappings to limit dictionary growth
            if shared_time_series:
                for ts_uuid in case_model.uuids(obj_type = 'TimeSeries'):
                    ts_title = case_model.title(uuid = ts_uuid)
                    ts_index = host_ts_titles.index(ts_title)
                    host_ts_uuid = host_ts_uuids[ts_index]
                    composite_model.force_consolidation_uuid_equivalence(ts_uuid, host_ts_uuid)
            if shared_grids:
                log.info('shared grids')  # debug
                for grid_uuid in case_model.uuids(obj_type = 'IjkGridRepresentation'):
                    grid_root = case_model.root(uuid = grid_uuid)
                    grid_extent = grr.extent_kji_from_root(grid_root)
                    host_index = None
                    if grid_extent in host_grid_shapes:
                        if title_match_required:
                            case_grid_title = rqet.citation_title_for_node(grid_root)
                            for host_grid_index in len(host_grid_uuids):
                                if grid_extent == host_grid_shapes[
                                        host_grid_index] and case_grid_title == host_grid_titles[host_grid_index]:
                                    host_index = host_grid_index
                                    break
                        else:
                            host_index = host_grid_shapes.index(grid_extent)
                    assert host_index is not None, 'failed to match grids when gathering ensemble'
                    composite_model.force_consolidation_uuid_equivalence(grid_uuid, host_grid_uuids[host_index])
                    grid_relatives = case_model.parts(related_uuid = grid_uuid)
                    t_props = 0.0
                    composite_h5_file_name = composite_model.h5_file_name()
                    composite_h5_uuid = composite_model.h5_uuid()
                    case_h5_file_name = case_model.h5_file_name()
                    for part in grid_relatives:
                        if 'Property' in part:
                            t_p_start = time()
                            composite_model.copy_part_from_other_model(case_model,
                                                                       part,
                                                                       realization = r,
                                                                       consolidate = True,
                                                                       force = shared_time_series,
                                                                       self_h5_file_name = composite_h5_file_name,
                                                                       h5_uuid = composite_h5_uuid,
                                                                       other_h5_file_name = case_h5_file_name)
                            t_props += time() - t_p_start
                    log.info(f'time props: {t_props:.3f} sec')  # debug
            else:
                log.info('non shared grids')  # debug
                composite_model.copy_all_parts_from_other_model(case_model, realization = r, consolidate = consolidate)
        log.info(f'case time: {time() - t_r_start:.2f} secs')  # debug

    if create_epc_lookup and len(epc_lookup_dict):
        epc_lookup = rqp.StringLookup(composite_model, int_to_str_dict = epc_lookup_dict, title = 'ensemble epc table')
        epc_lookup.create_xml()

    composite_model.store_epc()

    log.info(f'{len(epc_lookup_dict)} realizations merged into ensemble {new_epc_file}')



"""Function for creating a time series from a Nexus summary file."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import datetime as dt
import os
from ._time_series import TimeSeries


def time_series_from_nexus_summary(summary_file, parent_model = None):
    """Create a TimeSeries object based on time steps reported in a Nexus summary file (.sum).

    note:
        this function does not create the xml for the new TimeSeries, nor add it as a part to the parent model
    """

    if not summary_file:
        return None

    if not os.path.isfile(summary_file):
        log.warning('Summary file not found: ' + summary_file)
        return None

    try:

        summary_entries = _open_file(summary_file)

        time_series = _process_summary_entries(summary_entries, parent_model)

    except Exception:
        log.exception('failed to create TimeSeries object from summary file: ' + summary_file)
        return None

    return time_series


def _process_summary_entries(summary_entries, parent_model = None):
    """Create a TimeSeries object based on time steps reported in a Nexus summary file (.sum)."""
    if len(summary_entries) == 0:
        return None  # no entries extracted from summary file, could raise error?
    summary_entries.sort()
    if summary_entries[0][0] == 0:  # first entry is for time step zero
        tz_date = summary_entries[0][2]
        summary_entries.pop(0)
    else:  # back calculate time zero from first entry
        delta = dt.timedelta(days = summary_entries[0][1])
        tz_date = summary_entries[0][2] - delta
    time_series = TimeSeries(parent_model = parent_model, first_timestamp = tz_date.isoformat() + 'T00:00:00Z')
    last_timestep = 0
    last_cumulative_time = 0.0
    for entry in summary_entries:
        if entry[0] <= last_timestep:
            log.warning('ignoring out of sequence time step in summary file: ' + str(entry[0]))
            continue
        while entry[0] > last_timestep + 1:  # add filler steps
            time_series.extend_by_days(0)
            last_timestep += 1
        if entry[1] < last_cumulative_time:  # should never happen
            time_series.extend_by_days(0)
        else:
            time_series.extend_by_days(entry[1] - last_cumulative_time)
            last_cumulative_time = entry[1]
        last_timestep += 1
    return time_series


def _open_file(summary_file):
    """Opens the Nexus summary file."""

    us_date_format = True
    summary_entries = []  # list of (time step no., cumulative time (days), date (dd/mm/yyyy))

    with open(summary_file, "r") as fp:
        while True:
            line = fp.readline()
            if len(line) == 0:
                break  # end of file
            words = line.split()
            if len(words) < 5:
                continue
            if not words[0].isdigit():
                if words[0].upper() == 'NO.' and len(words[4]) == 10:
                    date_format = words[4].upper()
                    us_date_format = (date_format == 'MM/DD/YYYY')
                    if not us_date_format:
                        assert date_format == 'DD/MM/YYYY'
                continue
            ts_number = int(words[0])
            time_in_days = float(words[3])
            date_str = words[4]
            assert 9 <= len(date_str) <= 10, 'invalid date string length in summary file: ' + date_str
            if us_date_format:
                date = dt.date(int(date_str[-4:]), int(date_str[:-8]), int(date_str[-7:-5]))
            else:
                date = dt.date(int(date_str[-4:]), int(date_str[-7:-5]), int(date_str[:-8]))
            summary_entries.append((ts_number, time_in_days, date))
    return summary_entries



"""_binary_contact_interpretation.py: RESQML low level BinaryContactInterpretation class."""

version = '24th November 2021'

# NB: in this module, the term 'unit' refers to a geological stratigraphic unit, i.e. a layer of rock, not a unit of measure

import logging

log = logging.getLogger(__name__)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.xml_namespaces import curly_namespace as ns
from resqpy.strata._strata_common import valid_contact_relationships, valid_contact_verbs, valid_contact_sides, valid_contact_modes


class BinaryContactInterpretation:
    """Internal class for contact between 2 geological entities; not a high level class but used by others."""

    def __init__(
            self,
            model,
            existing_xml_node = None,
            index = None,
            contact_relationship: str = None,
            verb: str = None,
            subject_uuid = None,
            direct_object_uuid = None,
            subject_contact_side = None,  # optional
            subject_contact_mode = None,  # optional
            direct_object_contact_side = None,  # optional
            direct_object_contact_mode = None,  # optional
            part_of_uuid = None):  # optional
        """Creates a new binary contact interpretation internal object.

        note:
           if an existing xml node is present, then all the later arguments are ignored
        """
        # index (non-negative integer, should increase with decreasing age for horizon contacts)
        # contact relationship (one of valid_contact_relationships)
        # subject (reference to e.g. stratigraphic unit interpretation)
        # verb (one of valid_contact_verbs)
        # direct object (reference to e.g. stratigraphic unit interpretation)
        # contact sides and modes (optional)
        # part of (reference to e.g. horizon interpretation, optional)

        self.model = model

        if existing_xml_node is not None:
            self._load_from_xml(existing_xml_node)

        else:
            assert index >= 0
            assert contact_relationship in valid_contact_relationships
            assert verb in valid_contact_verbs
            assert subject_uuid is not None and direct_object_uuid is not None
            if subject_contact_side is not None:
                assert subject_contact_side in valid_contact_sides
            if subject_contact_mode is not None:
                assert subject_contact_mode in valid_contact_modes
            if direct_object_contact_side is not None:
                assert direct_object_contact_side in valid_contact_sides
            if direct_object_contact_mode is not None:
                assert direct_object_contact_mode in valid_contact_modes
            self.index = index
            self.contact_relationship = contact_relationship
            self.verb = verb
            self.subject_uuid = subject_uuid
            self.direct_object_uuid = direct_object_uuid
            self.subject_contact_side = subject_contact_side
            self.subject_contact_mode = subject_contact_mode
            self.direct_object_contact_side = direct_object_contact_side
            self.direct_object_contact_mode = direct_object_contact_mode
            self.part_of_uuid = part_of_uuid

    def _load_from_xml(self, bci_node):
        """Populates this binary contact interpretation based on existing xml.

        arguments:
           bci_node (lxml.etree._Element): the root xml node for the binary contact interpretation sub-tree
        """

        assert bci_node is not None

        self.contact_relationship = rqet.find_tag_text(bci_node, 'ContactRelationship')
        assert self.contact_relationship in valid_contact_relationships,  \
           f'missing or invalid contact relationship {self.contact_relationship} in xml for binary contact interpretation'

        self.index = rqet.find_tag_int(bci_node, 'Index')
        assert self.index is not None, 'missing index in xml for binary contact interpretation'

        self.part_of_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(bci_node, ['PartOf', 'UUID']))

        sr_node = rqet.find_tag(bci_node, 'Subject')
        assert sr_node is not None, 'missing subject in xml for binary contact interpretation'
        self.subject_uuid = bu.uuid_from_string(rqet.find_tag_text(sr_node, 'UUID'))
        assert self.subject_uuid is not None
        self.subject_contact_side = rqet.find_tag_text(sr_node, 'Qualifier')
        self.subject_contact_mode = rqet.find_tag_text(sr_node, 'SecondaryQualifier')

        dor_node = rqet.find_tag(bci_node, 'DirectObject')
        assert dor_node is not None, 'missing direct object in xml for binary contact interpretation'
        self.direct_object_uuid = bu.uuid_from_string(rqet.find_tag_text(dor_node, 'UUID'))
        assert self.direct_object_uuid is not None
        self.direct_object_contact_side = rqet.find_tag_text(dor_node, 'Qualifier')
        self.direct_object_contact_mode = rqet.find_tag_text(dor_node, 'SecondaryQualifier')

        self.verb = rqet.find_tag_text(bci_node, 'Verb')
        assert self.verb in valid_contact_verbs,  \
           f'missing or invalid contact verb {self.verb} in xml for binary contact interpretation'

    def create_xml(self, parent_node = None):
        """Generates xml sub-tree for this contact interpretation, for inclusion as element of high level interpretation.

        arguments:
           parent_node (lxml.etree._Element, optional): if present, the created sub-tree is added as a child to this node

        returns:
           lxml.etree._Element: the root node of the newly created xml sub-tree for the contact interpretation
        """

        bci = rqet.Element(ns['resqml2'] + 'ContactInterpretation')
        bci.set(ns['xsi'] + 'type', ns['resqml2'] + 'BinaryContactInterpretationPart')
        bci.text = ''

        cr_node = rqet.SubElement(bci, ns['resqml2'] + 'ContactRelationship')
        cr_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ContactRelationship')
        cr_node.text = self.contact_relationship

        i_node = rqet.SubElement(bci, ns['resqml2'] + 'Index')
        i_node.set(ns['xsi'] + 'type', ns['xsi'] + 'nonNegativeInteger')
        i_node.text = str(self.index)

        if self.part_of_uuid is not None:
            self.model.create_ref_node('PartOf',
                                       self.model.title(uuid = self.part_of_uuid),
                                       self.part_of_uuid,
                                       content_type = self.model.type_of_uuid(self.part_of_uuid),
                                       root = bci)

        dor_node = self.model.create_ref_node('DirectObject',
                                              self.model.title(uuid = self.direct_object_uuid),
                                              self.direct_object_uuid,
                                              content_type = self.model.type_of_uuid(self.direct_object_uuid),
                                              root = bci)
        dor_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ContactElementReference')

        if self.direct_object_contact_side:
            doq_node = rqet.SubElement(dor_node, ns['resqml2'] + 'Qualifier')
            doq_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ContactSide')
            doq_node.text = self.direct_object_contact_side

        if self.direct_object_contact_mode:
            dosq_node = rqet.SubElement(dor_node, ns['resqml2'] + 'SecondaryQualifier')
            dosq_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ContactMode')
            dosq_node.text = self.direct_object_contact_mode

        v_node = rqet.SubElement(bci, ns['resqml2'] + 'Verb')
        v_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ContactVerb')
        v_node.text = self.verb

        sr_node = self.model.create_ref_node('Subject',
                                             self.model.title(uuid = self.subject_uuid),
                                             self.subject_uuid,
                                             content_type = self.model.type_of_uuid(self.subject_uuid),
                                             root = bci)
        sr_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ContactElementReference')

        if self.subject_contact_side:
            sq_node = rqet.SubElement(sr_node, ns['resqml2'] + 'Qualifier')
            sq_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ContactSide')
            sq_node.text = self.subject_contact_side

        if self.subject_contact_mode:
            ssq_node = rqet.SubElement(sr_node, ns['resqml2'] + 'SecondaryQualifier')
            ssq_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ContactMode')
            ssq_node.text = self.subject_contact_mode

        if parent_node is not None:
            parent_node.append(bci)

        return bci



"""fine_coarse.py: Module providing support for grid refinement and coarsening."""

version = '20th October 2021'

# Nexus is a registered trademark of the Halliburton Company

# todo: cater for parallel to top & parallel to base style refinement proportions in k

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.box_utilities as box


class FineCoarse:
    """Class for holding a mapping between fine and coarse grids."""

    def __init__(self, fine_extent_kji, coarse_extent_kji, within_fine_box = None, within_coarse_box = None):
        """Partial initialisation function, call other methods to assign ratios and proportions.

        arguments:
           fine_extent_kji (triple int): the local extent of the fine grid in k, j, i axes, ie (nk, nj, ni).
           coarse_extent_kji (triple int): the local extent of the coarse grid in k, j, i axes.
           within_fine_box (numpy int array of shape (2, 3), optional): if present, the subset of a larger
              fine grid that the mapping refers to; axes are min,max and k,j,i; values are zero based indices;
              max values are included in box (un-pythonesque); use this in the case of a local grid coarsening
           within_coarse_box (numpy int array of shape (2, 3), optional): if present, the subset of a larger
              coarse grid that the mapping refers to; axes are min,max and k,j,i; values are zero based indices;
              max values are included in box (un-pythonesque); use this in the case of a local grid refinement;
              required for write_cartref() method to work

        returns:
           newly formed FineCoarse object awaiting determination of ratios and proportions by axes.

        notes:
           at most one of within_fine_box and within_coarse_box may be passed; this information is not really
           used internally by the FineCoarse class but is noted in order to support local grid refinement and
           local grid coarsening applications;
           after intialisation, set_* methods should be called to establish the mapping
        """

        assert len(fine_extent_kji) == 3 and len(coarse_extent_kji) == 3
        assert within_fine_box is None or within_coarse_box is None

        self.fine_extent_kji = tuple(fine_extent_kji)  #: fine extent
        self.coarse_extent_kji = tuple(coarse_extent_kji)  #: coarse extent
        self._assert_extents_valid()
        if within_fine_box is not None:
            _assert_valid_box(fine_extent_kji, within_fine_box)
        if within_coarse_box is not None:
            _assert_valid_box(coarse_extent_kji, within_coarse_box)
        self.within_fine_box = within_fine_box  #: if not None, a box within an unidentified larger fine grid
        self.within_coarse_box = within_coarse_box  #: if not None, a box within an unidentified larger coarse grid

        self.constant_ratios = [None, None, None]  #: list for 3 axes kji, each None or int
        self.vector_ratios = [None, None, None]  #: list for 3 axes kji, each numpy vector of int or None
        self.equal_proportions = [True, True,
                                  True]  #: list for 3 axes kji, each boolean defaulting to equal proportions
        self.vector_proportions = [None, None, None
                                  ]  #: list for 3 axes kji, each None or list of numpy vectors of float summing to 1.0

        self.fine_to_coarse_mapping = None  # derived triplet of int vectors holding coarse cell index for each fine cell index

    def assert_valid(self):
        """Checks consistency of everything within the fine coarse mapping; raises assertion error if not valid."""

        self._assert_extents_valid()

        assert self.constant_ratios is not None and len(self.constant_ratios) == 3
        assert self.vector_ratios is not None and len(self.vector_ratios) == 3
        assert self.equal_proportions is not None and len(self.equal_proportions) == 3
        assert self.vector_proportions is not None and len(self.vector_proportions) == 3

        for axis in range(3):
            assert (self.constant_ratios[axis] is None or
                    _is_int(self.constant_ratios[axis]) and self.constant_ratios[axis] > 0)
            if self.constant_ratios[axis] is None:
                assert (self.vector_ratios[axis] is not None and isinstance(self.vector_ratios[axis], np.ndarray) and
                        self.vector_ratios[axis].ndim == 1 and str(self.vector_ratios[axis].dtype).startswith('int'))
                assert len(self.vector_ratios[axis]) == self.coarse_extent_kji[axis]
                assert np.all(self.vector_ratios[axis] > 0)
                assert np.sum(self.vector_ratios[axis]) == self.fine_extent_kji[axis]
            else:
                assert self.vector_ratios[axis] is None
                assert self.coarse_extent_kji[axis] * self.constant_ratios[axis] == self.fine_extent_kji[axis]
            assert isinstance(self.equal_proportions[axis], bool)
            if self.equal_proportions[axis]:
                assert self.vector_proportions[axis] is None
            else:
                assert (self.vector_proportions[axis] is not None and
                        isinstance(self.vector_proportions[axis], list) and  # could allow a tuple
                        len(self.vector_proportions[axis]) == self.coarse_extent_kji[axis])
                for c0 in range(self.coarse_extent_kji[axis]):
                    assert (isinstance(self.vector_proportions[axis][c0], np.ndarray) and
                            self.vector_proportions[axis][c0].ndim == 1 and
                            str(self.vector_proportions[axis][c0].dtype).startswith('float') and
                            len(self.vector_proportions[axis][c0]) == self.ratio(axis, c0))
                    assert np.min(self.vector_proportions[axis][c0]) > 0.0
                    assert abs(np.sum(self.vector_proportions[axis][c0]) - 1.0) < 1.0e-6

    def ratio(self, axis, c0):
        """Return fine:coarse ratio in given axis and coarse slice."""

        if self.constant_ratios[axis] is not None:
            return self.constant_ratios[axis]
        return self.vector_ratios[axis][c0]

    def ratios(self, c_kji0):
        """Return find:coarse ratios triplet for coarse cell."""

        return tuple(self.ratio(axis, c_kji0[axis]) for axis in range(3))

    def coarse_for_fine(self):
        """Returns triplet of numpy int vectors being the axial coarse cell indices for the axial fine cell indices."""

        if self.fine_to_coarse_mapping is None:
            self._set_fine_to_coarse_mapping()
        return self.fine_to_coarse_mapping

    def coarse_for_fine_kji0(self, fine_kji0):
        """Returns the index of the coarse cell which the given fine cell falls within."""

        if self.fine_to_coarse_mapping is None:
            self._set_fine_to_coarse_mapping()
        return (self.fine_to_coarse_mapping[0][fine_kji0[0]], self.fine_to_coarse_mapping[1][fine_kji0[1]],
                self.fine_to_coarse_mapping[2][fine_kji0[2]])

    def coarse_for_fine_axial(self, axis, f0):
        """Returns the index, for a single axis, of the coarse cell which the given fine cell falls within."""

        if self.fine_to_coarse_mapping is None:
            self._set_fine_to_coarse_mapping()
        return self.fine_to_coarse_mapping[axis][f0]

    def coarse_for_fine_axial_vector(self, axis):
        """Returns a numpy int vector, for a single axis, of the coarse cell index which each fine cell falls within."""

        if self.fine_to_coarse_mapping is None:
            self._set_fine_to_coarse_mapping()
        return self.fine_to_coarse_mapping[axis]

    def fine_base_for_coarse_axial(self, axis, c0):
        """Returns the index, for a single axis, of the 'first' fine cell within the coarse cell (lowest fine index)."""

        if self.constant_ratios[axis] is not None:
            return c0 * self.constant_ratios[axis]
        return np.sum(self.vector_ratios[axis][:c0])  # todo: check this returns zero for c0 == 0

    def fine_base_for_coarse(self, c_kji0):
        """Returns a 3-tuple being the 'first' (min) k, j, i0 in fine grid for given coarse cell."""

        base = []
        for axis in range(3):
            base.append(self.fine_base_for_coarse_axial(axis, c_kji0[axis]))
        return tuple(base)

    def fine_box_for_coarse(self, c_kji0):
        """Return the min, max for k, j, i0 in fine grid for given coarse cell.
        
        Returns:
            Numpy int array of shape (2, 3) being the min, max for k, j, i0
        """

        box = np.empty((2, 3), dtype = int)
        box[0] = self.fine_base_for_coarse(c_kji0)
        box[1] = box[0] + self.ratios(c_kji0) - 1
        return box

    def proportion(self, axis, c0):
        """Return the axial relative proportions of fine within coarse.
        
        Returns:
            numpy vector of floats, summing to one
        """

        if self.equal_proportions[axis]:
            count = self.ratio(axis, c0)
            fraction = 1.0 / float(count)
            return np.full((count,), fraction)
        return self.vector_proportions[axis][c0]

    def proportions_for_axis(self, axis):
        """Return the axial relative proportions as array of floats summing to one."""

        if self.equal_proportions[axis]:
            count = self.constant_ratios[axis]
            fraction = 1.0 / float(count)
            return np.full((self.fine_extent_kji[axis],), fraction)
        return np.concatenate(self.vector_proportions[axis])

    def interpolation(self, axis, c0):
        """Return a float array ready for interpoltion.
        
        Returns floats starting at zero and increasing monotonically to less than one.
        """
        count = self.ratio(axis, c0)
        fractions = np.zeros((count,))
        proport = self.proportion(axis, c0)
        for f0 in range(1, count):
            fractions[f0] = fractions[f0 - 1] + proport[f0 - 1]
        return fractions

    def proportions(self, c_kji0):
        """Return triplet of axial proportions for refinement of coarse cell."""

        return (self.proportion(axis, c_kji0[axis]) for axis in range(3))

    def set_constant_ratio(self, axis):
        """Set the refinement ratio for axis based on the ratio of the fine to coarse extents."""

        assert 0 <= axis < 3
        extent_ratio, remainder = divmod(self.fine_extent_kji[axis], self.coarse_extent_kji[axis])
        assert remainder == 0, 'coarse extent in ' + 'KJI'[axis] + ' is not a multiple of fine extent'
        self.constant_ratios[axis] = extent_ratio
        self.vector_ratios[axis] = None

    def set_ij_ratios_constant(self):
        """Set the refinement ratio for I & J axes based on the ratio of the fine to coarse extents."""

        for axis in (1, 2):
            self.set_constant_ratio(axis)

    def set_all_ratios_constant(self):
        """Set all refinement ratios constant based on the ratio of the fine to coarse extents."""

        for axis in range(3):
            self.set_constant_ratio(axis)

    def set_ratio_vector(self, axis, vector):
        """Set fine:coarse ratios for axis from numpy int vector of length matching coarse extent."""

        assert 0 <= axis < 3
        if isinstance(vector, list) or isinstance(vector, tuple):
            vector = np.array(vector)
        assert isinstance(vector, np.ndarray) and vector.ndim == 1
        assert str(vector.dtype).startswith('int')
        assert len(vector) == self.coarse_extent_kji[axis],  \
               'length of vector of refinement ratios does not match coarse extent'
        assert np.sum(vector) == self.fine_extent_kji[axis],  \
               'sum of refinement ratios in vector does not match fine extent'
        minimum = np.min(vector)
        assert minimum > 0
        if minimum == np.max(vector):
            self.set_constant_ratio(axis)
            assert self.constant_ratios[axis] == minimum
        else:
            self.vector_ratios[axis] = vector.copy()
            self.constant_ratios[axis] = None

    def set_equal_proportions(self, axis):
        """Set proportions equal for axis."""

        assert 0 <= axis < 3
        self.equal_proportions[axis] = True
        self.vector_proportions[axis] = None

    def set_all_proprtions_equal(self):
        """Sets proportions equal in all 3 axes."""

        for axis in range(3):
            self.set_equal_proportions(axis)

    def set_proportions_list_of_vectors(self, axis, list_of_vectors):
        """Sets the proportions for given axis, with one vector for each coarse slice in the axis."""

        assert 0 <= axis < 3
        assert len(list_of_vectors) == self.coarse_extent_kji[axis]
        if self.vector_proportions is None:
            self.vector_proportions = [None, None, None]
        fractions_list_of_vectors = []
        all_equal = True
        for c0 in range(self.coarse_extent_kji[axis]):
            vector = list_of_vectors[c0]
            count = self.ratio(axis, c0)
            if isinstance(vector, list) or isinstance(vector, tuple):
                vector = np.array(vector)
            assert isinstance(vector, np.ndarray) and vector.ndim == 1
            assert len(vector) == count, 'wrong number of proportions for axis: ' + str(axis) + ' slice(0): ' + c0
            assert not np.any(np.isnan(vector))
            assert np.all(vector > 0)
            total = float(np.sum(vector))
            fractions = np.empty((count,), dtype = float)
            fractions[:] = vector
            fractions /= total
            fractions_list_of_vectors.append(fractions)
            if np.max(fractions) - np.min(fractions) > 1.0e-6:
                all_equal = False
        if all_equal:
            self.equal_proportions[axis] = True
            self.vector_proportions[axis] = None
        else:
            self.equal_proportions[axis] = False
            self.vector_proportions[axis] = fractions_list_of_vectors

    def fine_for_coarse_natural_column_index(self, coarse_col):
        """Returns the fine equivalent natural (first) column index for coarse natural column index."""

        j, i = divmod(coarse_col, self.coarse_extent_kji[2])
        ratio_j, ratio_i = self.constant_ratios[1], self.constant_ratios[2]
        if ratio_j is None:
            if j:
                j = np.sum(self.vector_ratios[1][0:j])
        else:
            j *= ratio_j
        if ratio_i is None:
            if i:
                i = np.sum(self.vector_ratios[2][0:i])
        else:
            i *= ratio_i
        return j * self.fine_extent_kji[2] + i

    def fine_for_coarse_natural_pillar_index(self, coarse_p):
        """Returns the fine equivalent natural (first) pillar index for coarse natural pillar index."""

        p_j, p_i = divmod(coarse_p, self.coarse_extent_kji[2] + 1)
        ratio_j, ratio_i = self.constant_ratios[1], self.constant_ratios[2]
        if ratio_j is None:
            if p_j:
                p_j = np.sum(self.vector_ratios[1][0:p_j])
        else:
            p_j *= ratio_j
        if ratio_i is None:
            if p_i:
                p_i = np.sum(self.vector_ratios[2][0:p_i])
        else:
            p_i *= ratio_i
        return p_j * (self.fine_extent_kji[2] + 1) + p_i

    def write_cartref(self,
                      filename,
                      lgr_name,
                      mode = 'a',
                      root_name = None,
                      preceeding_blank_lines = 0,
                      trailing_blank_lines = 0):
        """Write Nexus ascii input format CARTREF; within_coarse_box must have been set."""

        assert self.within_coarse_box is not None, 'box within larger coarse grid required for cartref'
        assert filename and lgr_name, 'filename or lgr name missing for cartref'
        assert ' ' not in lgr_name, 'lgr name may not contain spaces'
        if len(lgr_name) > 8 or (root_name and len(root_name) > 8):
            log.warning('Nexus might only differentiate first 8 characters of grid names')
        with open(filename, mode) as fp:
            for _ in range(preceeding_blank_lines):
                fp.write('\n')
            if root_name:
                fp.write('LGR ' + str(root_name) + '\n')
            fp.write('LGR\n')
            fp.write('CARTREF ' + str(lgr_name) + '\n')
            fp.write('   ' + box.spaced_string_iijjkk1_for_box_kji0(self.within_coarse_box) + '\n')
            for axis in range(2, -1, -1):
                fp.write('  ')
                for c0 in range(self.coarse_extent_kji[axis]):
                    fp.write(' ' + str(self.ratio(axis, c0)))
                fp.write('\n')
                if not self.equal_proportions[axis]:
                    log.warning('unequal propertions in axis ' + 'KJI'[axis] + ': define Lgr corp separately')
            fp.write('ENDREF\n')
            fp.write('ENDLGR\n')
            for _ in range(trailing_blank_lines):
                fp.write('\n')

    def _assert_extents_valid(self):

        for axis in range(3):
            assert _is_int(self.fine_extent_kji[axis]) and _is_int(self.coarse_extent_kji[axis])
            assert self.fine_extent_kji[axis] > 0
            assert self.coarse_extent_kji[axis] <= self.fine_extent_kji[axis]

    def _set_fine_to_coarse_mapping(self):
        self.fine_to_coarse_mapping = [None, None, None]
        for axis in range(3):
            self.fine_to_coarse_mapping[axis] = np.empty(self.fine_extent_kji[axis], dtype = int)
            f0 = 0
            for c0 in range(self.coarse_extent_kji[axis]):
                ratio = self.ratio(axis, c0)
                self.fine_to_coarse_mapping[axis][f0:f0 + ratio] = c0
                f0 += ratio
            assert f0 == self.fine_extent_kji[axis]


def tartan_refinement(coarse_extent_kji,
                      coarse_fovea_box,
                      fovea_ratios_kji,
                      decay_rates_kji = None,
                      decay_mode = 'exponential',
                      within_coarse_box = None):
    """Returns a new FineCoarse object set to a tartan grid refinement; fine extent is determined from arguments.

    arguments:
       coarse_extent_kji (triple int): the extent of the coarse grid being refined
       coarse_fovea_box (numpy int array of shape (2, 3)): the central box within the coarse grid to receive maximum refinement
       fovea_ratios_kji (triple int): the maximum refinement ratios, to be applied in the coarse_fovea_box
       decay_rates_kji (triple float or triple int, optional): controls how quickly refinement ratio reduces in slices away from fovea;
          if None then default values will be generated; see notes for more details
       decay_mode (str, default 'exponential'): 'exponential' or 'linear'; see notes
       within_coarse_box (numpy int array of shape (2, 3), optional): if present, is preserved in FineCoarse for possible use
          in setting resqml ParentWindow or generating Nexus CARTREF

    returns:
       FineCoarse object holding the tartan refinement mapping

    notes:
       each axis is treated independently;
       the fovea (box of maximum refinement) may be a column of cells (for a vertical well) or any other logical cuboid;
       the refinement factor is reduced monotonically in slices moving away from the fovea;
       two refinement factor decay functions are available: 'exponential' and 'linear', with different meaning to decay_rates_kji;
       for exponential decay, each decay rate should be a float in the range 0.0 to 1.0, with 0.0 causing immediate change
       to no refinement (factor 1), and 1.0 causing no decay (constant refinement at fovea factor);
       for linear decay, each decay rate should typically be a non-negative integer (though float is also okay), with 0 causing
       no decay, 1 causing a reduction in refinement factor of 1 per coarse slice, 2 meaning refinement factor reduces by 2 with
       each coarse slice etc.;
       in all cases, the refinement factor is given a lower limit of 1;
       the factor is rounded to an int for each slice, when working with floats;
       if decay rates are not passed as arguments, suitable values are generated to give a gradual reduction in refinement to a
       ratio of one at the boundary of the grid
    """

    assert box.valid_box(coarse_fovea_box, coarse_extent_kji)
    assert decay_mode in ['exponential', 'linear']
    assert len(fovea_ratios_kji) == 3

    fovea_ratios_kji = np.array(fovea_ratios_kji)
    assert np.all(fovea_ratios_kji >= 1)

    # generate decay rates if not passed in
    rates = np.zeros((2, 3), dtype = float)  # -/+ side of fovea, in each axis
    if decay_rates_kji is None:
        border = np.zeros((2, 3), dtype = float)
        border[0] = coarse_fovea_box[0].astype(float) + 0.5
        border[1] = (coarse_extent_kji - coarse_fovea_box[1]).astype(float) - 0.5
        for side in (0, 1):
            if decay_mode == 'linear':
                rates[side] = (fovea_ratios_kji - 1) / border[side]
            else:  # exponential decay
                rates[side] = np.power(1.5 / fovea_ratios_kji, 1.0 / border[side])
    else:
        assert len(decay_rates_kji) == 3
        rates[:] = decay_rates_kji

    log.debug(f'decay rates: {rates}')

    ratios_list = []
    fine_extent_kji = np.zeros(3, dtype = int)

    for axis in range(3):
        ratios = np.ones(coarse_extent_kji[axis], dtype = int)
        slice_a, slice_b = coarse_fovea_box[:, axis]
        ratios[slice_a:slice_b + 1] = fovea_ratios_kji[axis]
        floating_ratio = np.ones(2, dtype = float)
        if decay_mode == 'linear':
            floating_ratio[:] = fovea_ratios_kji[axis] - rates[:, axis]
        else:
            floating_ratio[:] = fovea_ratios_kji[axis] * rates[:, axis]  # exponential decay
        rounded_ratio = np.around(floating_ratio).astype(int)
        while np.any(rounded_ratio > 1) and (slice_a > 0 or slice_b < coarse_extent_kji[axis] - 1):
            slice_a -= 1
            slice_b += 1
            if slice_a >= 0:
                ratios[slice_a] = rounded_ratio[0]
            if slice_b < coarse_extent_kji[axis]:
                ratios[slice_b] = rounded_ratio[1]
            if decay_mode == 'linear':
                floating_ratio[:] -= rates[:, axis]
            else:
                floating_ratio[:] *= rates[:, axis]  # exponential decay
            rounded_ratio = np.around(floating_ratio).astype(int)
        ratios_list.append(ratios)
        fine_extent_kji[axis] = np.sum(ratios)


#      log.debug(f'retio vector [{axis}]: {ratios}')

    fc = FineCoarse(fine_extent_kji, coarse_extent_kji, within_coarse_box = within_coarse_box)

    for axis in range(3):
        fc.set_ratio_vector(axis, ratios_list[axis])
        # todo: set proportions ?

    return fc


def axis_for_letter(letter):
    """Returns 0, 1 or 2 for 'K', 'J', or 'I'; as required for axis arguments in FineCoarse methods."""

    assert isinstance(letter, str) and len(letter) == 1
    u = letter.upper()
    return 'KJI'.index(u)


def letter_for_axis(axis):
    """Returns K, J, or I for axis; axis as required for axis arguments in FineCoarse methods."""

    assert isinstance(axis, int) and 0 <= axis < 3
    return 'KJI'[axis]


def _assert_valid_box(extent_kji, box):
    assert isinstance(box, np.ndarray) and box.shape == (2, 3) and str(box.dtype).startswith('int')
    for axis in range(3):
        assert 0 <= box[0, axis] <= box[1, axis]
        assert box[1, axis] - box[0, axis] + 1 == extent_kji[axis]


def _is_int(v):
    t = type(v)
    if t is int:
        return True
    if t in [np.int64, np.int32, np.int16, np.int8]:
        return True
    s = str(t)
    if s.startswith('int'):
        return True
    return False



"""zone_layer_ranges_from_array() function."""

import logging

log = logging.getLogger(__name__)

import numpy as np


def zone_layer_ranges_from_array(zone_array, min_k0 = 0, max_k0 = None, use_dominant_zone = False):
    """Returns a list of (zone_min_k0, zone_max_k0, zone_index) derived from zone_array.

    arguments:
       zone_array (3D numpy int array or masked array): array holding zone index value per cell
       min_k0 (int, default 0): the minimum layer number (0 based) to be included in the ranges
       max_k0 (int, default None): the maximum layer number (0 based) to be included in the ranges;
          note that this layer is included (unlike in python ranges); if None, the maximum layer
          number in zone_array is used
       use_dominant_zone (boolean, default False): if True, the most common zone value in each layer is used for the whole
          layer; if False, then variation of zone values in active cells in a layer will raise an assertion error

    returns:
       a list of (int, int, int) being (zone_min_k0, zone_max_k0, zone_index) for each zone index value present

    notes:
       the function requires zone indices (for active cells, if zone_array is masked) within a layer
       to be consistent: an assertion error is raised otherwise; the returned list is sorted by layer
       ranges rather than zone index; if use_dominant_zone is True then a side effect of the function is
       to modify the values in zone_array to be consistent across each layer, effectively reassigning some
       cells to a different zone!
    """

    if max_k0 is None:
        max_k0 = zone_array.shape[0] - 1
    if use_dominant_zone:
        _dominant_zone(zone_array)
    zone_list = np.unique(zone_array)
    log.debug('list of zones: ' + str(zone_list))
    assert len(zone_list) > 0, 'no zone values present (all cells inactive?)'
    zone_layer_range_list = []  # list of (zone_min_k0, zone_max_k0, zone)
    for zone in zone_list:
        single_zone_mask = (zone_array == zone)
        zone_min_k0 = None
        zone_max_k0 = None
        for k0 in range(min_k0, max_k0 + 1):
            if np.any(single_zone_mask[k0]):
                zone_min_k0 = k0
                break
        if zone_min_k0 is None:
            log.warning('no active cells for zone ' + str(zone) + ' in layer range ' + str(min_k0 + 1) + ' to ' +
                        str(max_k0 + 1))
            continue
        for k0 in range(max_k0, min_k0 - 1, -1):
            if np.any(single_zone_mask[k0]):
                zone_max_k0 = k0
                break
        # require all active cells in zone layer range to be for this zone
        for k0 in range(zone_min_k0, zone_max_k0 + 1):
            assert np.all(single_zone_mask[k0]), 'unacceptable zone variation with layer ' + str(k0 + 1)
        zone_layer_range_list.append((zone_min_k0, zone_max_k0, zone))
    assert len(zone_layer_range_list) > 0, 'no zone layer ranges derived from zone array'
    zone_layer_range_list.sort()
    for zone_i in range(1, len(zone_layer_range_list)):
        assert zone_layer_range_list[zone_i][0] > zone_layer_range_list[
            zone_i - 1][1], 'overlapping zone layer ranges'  # todo: add more info
        if zone_layer_range_list[zone_i][0] > zone_layer_range_list[zone_i - 1][1] + 1:
            log.warning('gap in zonal layer ranges, missing layer(s) being arbitrarily assigned to zone below'
                       )  # todo: add more info to log
            zone_layer_range_list[zone_i][0] = zone_layer_range_list[zone_i - 1][1] + 1
    return zone_layer_range_list


def _dominant_zone(zone_array):
    # modifies data in zone_array such that each layer has a single (most common) value
    for k in range(zone_array.shape[0]):
        unique_zones = np.unique(zone_array[k])
        if len(unique_zones) <= 1:
            continue
        dominant = unique_zones[0]
        dominant_count = np.count_nonzero(zone_array[k] == dominant)
        for contender in unique_zones[1:]:
            contender_count = np.count_nonzero(zone_array[k] == contender)
            if contender_count > dominant_count:
                dominant = contender
                dominant_count = contender_count
        zone_array[k] = dominant
        log.info('layer ' + str(k + 1) + ' (1 based) zone set to most common value ' + str(dominant))



""" trademark.py module for mentioning trademarks in diagnostic log """

version = '9th April 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging as lg

log = lg.getLogger(__name__)

nexus_tm_level = None


def log_nexus_tm(level = lg.INFO):
    """Produces a Nexus trademark log message once at the given severity.

    note:
       this function should be called after referring to Nexus in another log message, passing the
       severity of that other message
    """

    global nexus_tm_level
    if isinstance(level, str):
        level = lg.__dict__[level.upper()]
    if nexus_tm_level is None or level > nexus_tm_level:
        preamble = '(ignore severity) ' if level > 20 else ''
        log.log(level, preamble + 'Nexus is a registered trademark of the Halliburton Company')
        nexus_tm_level = level



"""xml_et.py: Resqml xml element tree utilities module."""

version = '30th November 2021'

import logging

log = logging.getLogger(__name__)

import os

# import xml element tree parse method and classes here to allow single point for switching between lxml and etree
# alternative to lxml.etree: xml.etree.ElementTree
from lxml.etree import (  # type: ignore
    Element, ElementTree, SubElement, _Element,  # noqa
    parse)

import resqpy.olio.uuid as bu
from resqpy.olio.xml_namespaces import curly_namespace as cns
from resqpy.olio.xml_namespaces import inverse_namespace as inv_ns
from resqpy.olio.xml_namespaces import namespace as ns

pretend_to_be_fesapi = False
use_fesapi_quirks = True
use_tabs = False

if use_fesapi_quirks:
    null_xml_text = '\n'
else:
    null_xml_text = ''


def strip_path(full_path):
    """Returns the filename part of full_path with any directory path removed.

    :meta private:
    """

    return os.path.basename(full_path)


def stripped_of_prefix(prefixed):
    """Returns a simplified version of an xml tag or other element with any {xsd defining prefix} stripped off."""

    if prefixed is None:
        return None
    if '}' in prefixed:
        return prefixed[prefixed.rfind('}') + 1:]
    return prefixed[prefixed.rfind(':') + 1:]


def colon_prefixed(curly_prefixed):
    """Returns a version of an xml tag with {url} prefix replaced with nsi: equivalent; also returns the nsi prefix."""

    if not curly_prefixed:
        return None, None
    if curly_prefixed[0] != '{':
        colon = curly_prefixed.find(':')
        if colon < 0:
            return curly_prefixed, None
        return curly_prefixed, curly_prefixed[:colon]
    pre_end = curly_prefixed.rfind('}')
    try:
        pre_colon = inv_ns[curly_prefixed[1:pre_end]]
    except Exception:
        return curly_prefixed, None
    return pre_colon + ':' + curly_prefixed[pre_end + 1:], pre_colon


def find_tag(root, tag_name, must_exist = False):
    """Finds the first child in xml node with a (prefix-stripped) tag matching given tag name."""

    if root is None:
        return None
    for child in root:
        if stripped_of_prefix(child.tag) == tag_name:
            return child
    if must_exist:
        raise ValueError(f"Expected tag {tag_name} not found in root {root}")
    return None


def find_tag_text(root, tag_name, must_exist = False):
    """Finds the first child in xml node with a tag matching given tag name; returns stripped text field."""

    if root is None:
        return None
    for child in root:
        if stripped_of_prefix(child.tag) == tag_name:
            return node_text(child)
    if must_exist:
        raise ValueError(f"Expected tag {tag_name} not found in root {root}")
    return None


def find_tag_bool(root, tag_name, must_exist = False):
    """Finds the first child in xml node with a tag matching given tag name; returns stripped text field as bool."""

    if root is None:
        return None
    for child in root:
        if stripped_of_prefix(child.tag) == tag_name:
            return node_bool(child)
    if must_exist:
        raise ValueError(f"Expected tag {tag_name} not found in root {root}")
    return None


def find_tag_int(root, tag_name, must_exist = False):
    """Finds the first child in xml node with a tag matching given tag name; returns stripped text field as int."""

    if root is None:
        return None
    for child in root:
        if stripped_of_prefix(child.tag) == tag_name:
            return node_int(child)
    if must_exist:
        raise ValueError(f"Expected tag {tag_name} not found in root {root}")
    return None


def find_tag_float(root, tag_name, must_exist = False):
    """Finds the first child in xml node with a tag matching given tag name; returns stripped text field as float."""

    if root is None:
        return None
    for child in root:
        if stripped_of_prefix(child.tag) == tag_name:
            return node_float(child)
    if must_exist:
        raise ValueError(f"Expected tag {tag_name} not found in root {root}")
    return None


def find_nested_tags(root, tag_list):
    """Follows a list of tags in a nested xml hierarchy, returning the node at the deepest level."""

    if not tag_list:
        return None
    head = find_tag(root, tag_list[0])
    if head is None:
        return None
    if len(tag_list) == 1:
        return head
    return find_nested_tags(head, tag_list[1:])


def find_nested_tags_cast(root, tag_list, dtype = None):
    """Return value of nested tags as desired dtype.

    Follows a list of tags in a nested xml hierarchy, returning the stripped text of the node at the deepest level.
    """

    cast_func = {
        int: node_int,
        float: node_float,
        bool: node_bool,
        str: node_text,
        None: lambda x: x,
    }[dtype]

    node = find_nested_tags(root, tag_list)
    return cast_func(node)


def find_nested_tags_text(root, tag_list):
    """Return stripped text of node at deepest level of xml hierarchy.
    
    Args:
        tag_list: list of tags in a nested xml hierarchy
    """

    node = find_nested_tags(root, tag_list)
    return node_text(node)


def find_nested_tags_bool(root, tag_list):
    """Return stripped text of node at deepest level of xml hierarchy as a bool.
    
    Args:
        tag_list: list of tags in a nested xml hierarchy
    """
    node = find_nested_tags(root, tag_list)
    return node_bool(node)


def find_nested_tags_int(root, tag_list):
    """Return stripped text of node at deepest level of xml hierarchy as an int.
    
    Args:
        tag_list: list of tags in a nested xml hierarchy
    """
    node = find_nested_tags(root, tag_list)
    return node_int(node)


def find_nested_tags_float(root, tag_list):
    """Return stripped text of node at deepest level of xml hierarchy as a float.
    
    Args:
        tag_list: list of tags in a nested xml hierarchy
    """
    node = find_nested_tags(root, tag_list)
    return node_float(node)


def count_tag(root, tag_name):
    """Returns the number of children in xml node with a (prefix-stripped) tag matching given tag name."""

    if root is None:
        return None
    count = 0
    for child in root:
        if stripped_of_prefix(child.tag) == tag_name:
            count += 1
    return count


def list_of_tag(root, tag_name):
    """Returns a list of children in xml node with a (prefix-stripped) tag matching given tag name."""

    if root is None:
        return None
    results = []
    for child in root:
        if stripped_of_prefix(child.tag) == tag_name:
            results.append(child)
    return results


def list_of_descendant_tag(root, tag_name):
    """Returns a list of descendants in xml node tree with a (prefix-stripped) tag matching given tag name."""

    if root is None:
        return None
    results = []
    for child in root.iterdescendants():
        if stripped_of_prefix(child.tag) == tag_name:
            results.append(child)
    return results


def list_obj_references(root, skip_hdf5 = True):
    """Returns list of nodes of type DataObjectReference."""

    if root is None:
        return None
    results = []
    if node_type(root) == 'DataObjectReference' and not (skip_hdf5 and stripped_of_prefix(root.tag) == 'HdfProxy'):
        results.append(root)
    for child in root:
        results += list_obj_references(child, skip_hdf5 = skip_hdf5)
    return results


def cut_obj_references(root, uuids_to_be_cut):
    """Deletes any object reference nodes to uuids in given list."""

    if root is None or not uuids_to_be_cut:
        return
    for child in root:
        if node_type(child) == 'DataObjectReference':
            referred_uuid = bu.uuid_from_string(find_tag_text(child, 'UUID', must_exist = True))
            for cut_uuid in uuids_to_be_cut:
                if bu.matching_uuids(referred_uuid, cut_uuid):
                    root.remove(child)
                    break
        else:
            cut_obj_references(child, uuids_to_be_cut)


def cut_nodes_of_types(root, types_to_be_cut):
    """Deletes any nodes of a type matching one in the given list."""

    if root is None or not types_to_be_cut:
        return
    for child in root:
        if node_type(child) in types_to_be_cut:
            root.remove(child)  # hope this doesn't mess up the iteration
        else:
            cut_nodes_of_types(child, types_to_be_cut)


def content_type(content_type_str):
    """Returns the actual type, as embedded in an xml ContentType attribute; application and version are disregarded."""

    if content_type_str is None:
        return None
    if 'type=' in content_type_str:
        return content_type_str[content_type_str.rfind('type=') + 5:]

    #   if ':' in content_type_str:
    #      return content_type_str[content_type_str.rfind(':') + 1:]
    return content_type_str


def node_type(node, is_rels = False, strip_obj = False):
    """Returns the type as held in attributes of xml node; defining authority is stripped out."""

    result = None
    if node is None:
        return None
    if is_rels:
        if 'Type' not in node.attrib.keys():
            return None
        type_str = node.attrib['Type']
        result = type_str[type_str.rfind('/') + 1:]
    else:
        for key in node.attrib.keys():
            if stripped_of_prefix(key) == 'type':
                #           type_str = node.attrib[key]
                #           return type_str[type_str.rfind(':') + 1:]
                type_str = stripped_of_prefix(node.attrib[key])
                result = type_str
    if result and strip_obj and result.startswith('obj_'):
        result = result[4:]
    return result


def print_xml_tree(root,
                   level = 0,
                   max_level = None,
                   strip_tag_refs = True,
                   to_log = False,
                   log_level = None,
                   max_lines = 0,
                   line_count = 0):
    """Print an xml tree in an indented semi-readable format; return accumulated number of lines."""

    if root is None or (max_level is not None and level > max_level):
        return line_count
    if log_level is not None:
        to_log = True
    if log_level is None or log_level == 'info':
        print_fn = log.info
    elif log_level == 'debug':
        print_fn = log.debug
    elif log_level in ['warn', 'warning']:
        print_fn = log.warning
    elif log_level == 'error':
        print_fn = log.error
    else:
        print_fn = log.critical
    if max_lines and line_count >= max_lines:
        if line_count == max_lines:
            message = '(...xml tree output truncated after ' + str(max_lines) + ' lines)'
            if to_log:
                print_fn(message)
            else:
                print(message)
            line_count += 1
        return line_count

    value = root.text
    type_attr = None
    if strip_tag_refs:
        tag = stripped_of_prefix(root.tag)
        attrib_dict = {}
        for key in root.attrib.keys():
            stripped_key = stripped_of_prefix(key)
            attrib_dict[stripped_key] = root.attrib[key]
            if stripped_key == 'type':
                type_attr = stripped_of_prefix(root.attrib[key])
    else:
        tag = root.tag
        attrib_dict = root.attrib
        for key in root.attrib.keys():
            if stripped_of_prefix(key) == 'type':
                type_attr = root.attrib[key]
    if to_log:
        message = (3 * level * ' ') + tag + ' # ' + str(attrib_dict)
        if type_attr is not None:
            message += ' ## ' + type_attr
        if value is not None:
            message += ' ### ' + root.text.replace('\n', '\\n')
        print_fn(message)
    else:
        print((3 * level * ' ') + tag, '#', attrib_dict, end = ' ')
        if type_attr is not None:
            print('##', type_attr, end = ' ')
        if value is not None:
            print('###', root.text.replace('\n', '\\n'), end = '')
        print('')
    line_count += 1
    for child in root:
        line_count = print_xml_tree(child,
                                    level = level + 1,
                                    max_level = max_level,
                                    strip_tag_refs = strip_tag_refs,
                                    to_log = to_log,
                                    log_level = log_level,
                                    max_lines = max_lines,
                                    line_count = line_count)
        if line_count > max_lines:
            break
    return line_count


def uuid_in_part_name(part_name):
    """Returns uuid as embedded in part name."""

    # This might not always work
    if part_name is None:
        return None
    if part_name.endswith('.xml') and len(part_name) >= 40:
        return bu.uuid_from_string(part_name[-40:-4])
    elif part_name.endswith('.xml.rels') and len(part_name) >= 45:
        return bu.uuid_from_string(part_name[-45:-9])
    return None


def part_name_for_object(obj_type, uuid, prefixed = False, epc_subdir = None):
    """Returns the standard part name comprised of the object type, uuid and .xml extension."""

    if prefixed and (pretend_to_be_fesapi or use_fesapi_quirks) and obj_type[0] != '/':
        prefix = '/'
    else:
        prefix = ''
    if not obj_type.startswith('obj_'):
        prefix += 'obj_'
    if epc_subdir:
        if not epc_subdir.endswith('/'):
            epc_subdir += '/'
        prefix = epc_subdir + prefix
    return prefix + obj_type + '_' + str(uuid) + '.xml'


def rels_part_name_for_part(part_name):
    """Returns the paired relationships part name for the given part name."""

    pn = stripped_of_prefix(part_name)
    if pn is None or len(pn) == 0:
        return None
    dir_place = pn.rfind('/')
    if dir_place == -1:
        return '_rels/' + pn + '.rels'
    if dir_place == 0:
        return '_rels' + pn + '.rels'
    if dir_place == len(pn) - 1:
        return None
    return pn[:dir_place + 1] + '_rels' + pn[dir_place:] + '.rels'


def uuid_for_part_root(root):
    """Returns uuid as stored in xml attribs for root."""

    if root is None:
        return None
    if 'uuid' not in root.attrib.keys():
        return None
    return bu.uuid_from_string(root.attrib['uuid'])


def patch_uuid_in_part_root(root, uuid):
    """Returns modified part name with uuid swapped to uuid argument; root attrib is also changed."""

    if root is None or uuid is None:
        return None
    # This might not always work
    root.attrib['uuid'] = str(uuid)
    return part_name_for_part_root(root)


def part_name_for_part_root(root, is_rels = False, epc_subdir = None):
    """Returns the part name given the root node for the part's xml."""

    if root is None:
        return None
    obj_type = node_type(root, is_rels = is_rels)
    uuid = uuid_for_part_root(root)
    if obj_type is None or uuid is None:
        return None
    return part_name_for_object(obj_type, uuid, epc_subdir = epc_subdir)


# the next two functions aren't really much to do with the xml element tree


def find_in_ordered_data(value, array_1d):
    """Returns the index in the ordered list-like array of value; or None if not present."""

    def find_in_subset(value, array_1d, start, end):
        # recursive binary split
        if start >= end:
            return None
        mid = start + (end - start) // 2
        sample = array_1d[mid]
        if sample == value:
            while mid > 0 and array_1d[mid - 1] == value:
                mid -= 1
            return mid
        if sample > value:
            return find_in_subset(value, array_1d, start, mid)
        if mid == start:
            mid += 1
        return find_in_subset(value, array_1d, mid, end)

    return find_in_subset(value, array_1d, 0, len(array_1d))


def simplified_data_type(array_dtype):
    """Returns a simplified string version of the elemental data type (typically for a numpy or hdf5 array)."""

    str_dtype = str(array_dtype)
    if str_dtype.startswith('int'):
        return 'int'
    if str_dtype.startswith('float') or str_dtype.startswith('real'):
        return 'float'
    if str_dtype.startswith('bool'):
        return 'bool'
    return str_dtype


# following functions mostly previously in resqml_print.py


def bool_from_text(text):
    """Returns boolean value for string 'true' or 'false'; anything else results in None."""

    if text is None:
        return None
    if text.strip().lower() == 'true':
        return True
    if text.strip().lower() == 'false':
        return False
    return None


def node_text(node, unknown_if_none = False):
    """Returns stripped node text or 'unknown' if node is None or text is blank or newline."""

    if node is None or node.text is None:
        return 'unknown' if unknown_if_none else None
    text = node.text.strip()
    if len(text):
        return text
    return 'unknown' if unknown_if_none else None


def node_bool(node):
    """Returns stripped node text as bool, or None."""

    if node is None:
        return None
    return bool_from_text(node.text)


def node_int(node):
    """Returns stripped node text as int, or None."""

    if node is None:
        return None
    text = node.text.strip()
    if text.lower() == 'none':
        return None
    if len(text):
        return int(text)
    return None


def node_float(node):
    """Returns stripped node text as float, or None."""

    if node is None:
        return None
    text = node.text.strip()
    if text.lower() == 'none':
        return None
    if len(text):
        return float(text)
    return None


def length_units_from_node(node):
    """Returns standard length units string based on node text, or 'unknown'."""

    if node is None or node.text == '' or node.text == '\n':
        return 'unknown'
    else:
        return node.text.strip()


def time_units_from_node(node):
    """Returns standard time units string based on node text, or 'unknown'."""

    if node is None or node.text == '' or node.text == '\n':
        return 'unknown'
    else:
        return node.text.strip()


def xyz_handedness(xy_axes: str, z_inc_down: bool):
    """Return xyz true handedness as 'left', 'right' or 'unknown'."""

    if xy_axes is None or z_inc_down is None:
        return 'unknown'
    xy_axes_split = xy_axes.lower().split()
    if len(xy_axes_split) != 2:
        return 'unknown'
    if xy_axes_split not in [
        ['easting', 'northing'],
        ['northing', 'easting'],  # only these 6 options allowed in resqml
        ['westing', 'southing'],
        ['southing', 'westing'],
        ['northing', 'westing'],
        ['westing', 'northing']
    ]:
        return 'unknown'
    right_handed = z_inc_down
    if xy_axes_split in [['easting', 'northing'], ['westing', 'southing'], ['northing', 'westing']]:
        right_handed = not right_handed
    if right_handed:
        return 'right'
    else:
        return 'left'


def ijk_handedness(geom_node):
    """Returns ijk true handedness as 'left', 'right' or 'unknown'.
    
    Args:
        GridIsRightHanded node in grid geometry node.
    """

    if geom_node is None:
        return 'unknown'
    right_handed = bool_from_text(node_text(find_tag(geom_node, 'GridIsRighthanded')))
    if right_handed is None:
        return 'unknown'
    if right_handed:
        return 'right'
    return 'left'


def citation_title_for_node(node):
    """Looks for a citation node as a child of node and returns the title text."""

    return find_nested_tags_text(node, ['Citation', 'Title'])


def creation_date_for_node(node):
    """Looks for a citation node as a child of node and returns the creation (date-time) text."""

    return find_nested_tags_text(node, ['Citation', 'Creation'])


def write_xml_node(xml_fp, root, level = 0, namespace_keys = []):
    """Recursively write an xml node to an open file; return number of nodes written."""

    def _escaped_text(text):
        d = {'<': '&lt;', '>': '&gt;', '&': '&amp;'}
        esc = '<&>'
        # todo: include quotes if needed
        e = ''
        for ch in str(text):
            if ch in esc:
                e += d[ch]
            else:
                e += ch
        return e

    if root is None:
        return 0

    ns_keys = namespace_keys.copy()

    type_attr = None
    tag, pre_colon = colon_prefixed(root.tag)

    if pre_colon in ['content_types', 'rels']:
        tag = tag[len(pre_colon) + 1:]
        ct_special = True
    else:
        ct_special = False

    if use_tabs:
        line = (level * '\t')
    else:
        line = (3 * level * ' ')
    line += '<' + tag  # todo: if any tags involve special characters, use _escaped_text(tag)
    if pre_colon and pre_colon not in ns_keys:
        line += ' xmlns'
        if not ct_special:
            line += ':' + pre_colon
        line += '="' + ns[pre_colon] + '"'
        ns_keys.append(pre_colon)

    attrib_ns_list = []
    attrib_list = []
    type_pre_colon = None
    for key in root.attrib.keys():
        colon_attrib_key, pre_colon_attrib = colon_prefixed(key)
        if pre_colon_attrib and pre_colon_attrib not in ns_keys:
            attrib_ns_list.append(pre_colon_attrib)
        if stripped_of_prefix(key) == 'type':
            type_attr, type_pre_colon = colon_prefixed(root.attrib[key])
            attrib_list.append(colon_attrib_key + '="' + type_attr + '"')
        elif ct_special and colon_attrib_key == 'PartName' and root.attrib[key].startswith('obj_'):
            attrib_list.append(colon_attrib_key + '="/' + root.attrib[key] + '"')
        else:
            attrib_list.append(colon_attrib_key + '="' + root.attrib[key] + '"')

    for attrib_ns in attrib_ns_list:
        line += ' xmlns:' + attrib_ns + '="' + ns[attrib_ns] + '"'
        ns_keys.append(attrib_ns)
    if type_pre_colon and type_pre_colon not in ns_keys:  # must be included in the local xml line?
        line += ' xmlns:' + type_pre_colon + '="' + ns[type_pre_colon] + '"'
        ns_keys.append(type_pre_colon)

    for attrib in attrib_list:
        line += ' ' + attrib

    node_count = 1

    if ct_special and len(root) == 0:

        line += '/>\n'
        xml_fp.write(line.encode())
    #      print(line, end = '') # debug

    else:

        line += '>'

        if root.text and not root.text.isspace():
            line += _escaped_text(root.text)

        xml_fp.write(line.encode())
        #      print(line, end = '') # debug
        indentation = ''

        if len(root):
            xml_fp.write(b'\n')
            #         print()  # debug
            if use_tabs:
                indentation = (level * '\t')
            else:
                indentation = (3 * level * ' ')
            for child in root:
                node_count += write_xml_node(xml_fp, child, level = level + 1, namespace_keys = ns_keys)

        line = indentation + '</' + tag + '>\n'
        xml_fp.write(line.encode())

    #      print(line, end = '') # debug

    return node_count


def write_xml(xml_fp, tree, standalone = None):
    """Write an xml tree to file in an indented format; gSOAP/FESAPI compatible; return number of nodes written."""

    #   print('-------------------------------------------------------------------------------------') # debug
    line = '<?xml version="1.0" encoding="UTF-8"'
    if standalone is not None:
        line += ' standalone="' + standalone + '"'
    line += '?>\n'
    xml_fp.write(line.encode())
    #   print(line, end = '')  # debug
    nodes = write_xml_node(xml_fp, tree.getroot())

    return nodes


def load_metadata_from_xml(node):
    """Loads the ExtraMetaData stored in a RESQML part as a dictionary."""

    if node is None:
        return None
    extra_metadata = {}
    meta_nodes = list_of_tag(node, 'ExtraMetadata')
    for meta in meta_nodes:
        name = find_tag_text(meta, 'Name')
        value = find_tag_text(meta, 'Value')
        extra_metadata[name] = value
    return extra_metadata


def create_metadata_xml(node, extra_metadata):
    """Writes the xml for the given metadata dictionary."""

    if extra_metadata:
        for data in extra_metadata.keys():
            metadata = SubElement(node, cns['resqml2'] + 'ExtraMetadata')
            metadata.set(cns['xsi'] + 'type', cns['resqml2'] + 'NameValuePair')
            metadata.text = null_xml_text

            name = SubElement(metadata, cns['resqml2'] + 'Name')
            name.set(cns['xsi'] + 'type', cns['xsd'] + 'string')
            name.text = str(data)

            value = SubElement(metadata, cns['resqml2'] + 'Value')
            value.set(cns['xsi'] + 'type', cns['xsd'] + 'string')
            value.text = str(extra_metadata[data])

    return node


def is_node(obj):
    """Returns True if type of object is element tree node; False otherwise."""

    # note: only tested for lxml

    return type(obj) is _Element or type(obj) is Element



""" Utilities for working with 3D vectors in cartesian space.

note: some of these functions are redundant as they are provided by built-in numpy operations.

a vector is a one dimensional numpy array with 3 elements: x, y, z.
some functions accept a tuple or list of 3 elements as an alternative to a numpy array.
"""

version = '8th March 2022'

import logging

log = logging.getLogger(__name__)

import math as maths
import numpy as np
import numba  # type: ignore
from numba import njit


def radians_from_degrees(deg):
    """Converts angle from degrees to radians."""
    return np.radians(deg)


def degrees_from_radians(rad):
    """Converts angle from radians to degrees."""
    return np.degrees(rad)


def zero_vector():
    """Returns a zero vector [0.0, 0.0, 0.0]."""
    return np.zeros(3)


def v_3d(v):
    """Returns a 3D vector for a 2D or 3D vector."""
    assert 2 <= len(v) <= 3
    if len(v) == 3:
        return v
    v3 = np.zeros(3)
    v3[:2] = v
    return v3


def add(a, b):  # note: could just use numpy a + b facility
    """Returns vector sum a+b."""
    a = np.array(a)
    b = np.array(b)
    assert a.size == b.size
    return a + b


def subtract(a, b):  # note: could just use numpy a - b facility
    """Returns vector difference a-b."""
    a = np.array(a)
    b = np.array(b)
    assert a.size == b.size
    return a - b


def elemental_multiply(a, b):  # note: could just use numpy a * b facility
    """Returns vector with products of corresponding elements of a and b."""
    a = np.array(a)
    b = np.array(b)
    assert a.size == b.size
    return a * b


def amplify(v, scaling):  # note: could just use numpy a * scalar facility
    """Returns vector with direction of v, amplified by scaling."""
    v = np.array(v)
    return scaling * v


def unit_vector(v):
    """Returns vector with same direction as v but with unit length."""
    assert 2 <= len(v) <= 3
    v = np.array(v, dtype = float)
    if np.all(v == 0.0):
        return v
    return v / maths.sqrt(np.sum(v * v))


def unit_vectors(v):
    """Returns vectors with same direction as those in v but with unit length."""
    scaling = np.sqrt(np.sum(v * v, axis = -1))
    zero_mask = np.zeros(v.shape, dtype = bool)
    zero_mask[np.where(scaling == 0.0), :] = True
    restore = np.seterr(all = 'ignore')
    result = np.where(zero_mask, 0.0, v / np.expand_dims(scaling, -1))
    np.seterr(**restore)
    return result


def nan_unit_vectors(v):
    """Returns vectors with same direction as those in v but with unit length, allowing NaNs."""
    nan_mask = np.isnan(v)
    restore = np.seterr(all = 'ignore')
    scaling = np.sqrt(np.sum(v * v, axis = -1))
    zero_mask = np.zeros(v.shape, dtype = bool)
    zero_mask[np.where(scaling == 0.0), :] = True
    result = np.where(zero_mask, 0.0, v / np.expand_dims(scaling, -1))
    result = np.where(nan_mask, np.nan, result)
    np.seterr(**restore)
    return result


def unit_vector_from_azimuth(azimuth):
    """Returns horizontal unit vector in compass bearing given by azimuth (x = East, y = North)."""
    azimuth = azimuth % 360.0
    azimuth_radians = radians_from_degrees(azimuth)
    result = zero_vector()
    result[0] = maths.sin(azimuth_radians)  # x (increasing to east)
    result[1] = maths.cos(azimuth_radians)  # y (increasing to north)
    return result  # leave z as zero


def azimuth(v):  # 'azimuth' is synonymous with 'compass bearing'
    """Returns the compass bearing in degrees of the direction of v (x = East, y = North), ignoring z."""
    assert 2 <= v.size <= 3
    z_zero_v = np.zeros(3)
    z_zero_v[:2] = v[:2]
    unit_v = unit_vector(z_zero_v)  # also checks that z_zero_v is not zero vector
    x = unit_v[0]
    y = unit_v[1]  # ignore z component
    if x == 0.0 and y == 0.0:
        return 0.0  # arbitrary azimuth of a vertical vector
    if abs(x) >= abs(y):
        radians = maths.pi / 2.0 - maths.atan(y / x)
        if x < 0.0:
            radians += maths.pi
    else:
        radians = maths.atan(x / y)
        if y < 0.0:
            radians += maths.pi
    if radians < 0.0:
        radians += 2.0 * maths.pi
    return degrees_from_radians(radians)


def azimuths(va):  # 'azimuth' is synonymous with 'compass bearing'
    """Returns the compass bearings in degrees of the direction of each vector in va (x = East, y = North), ignoring z."""
    assert va.ndim > 1 and 2 <= va.shape[-1] <= 3
    shape = tuple(list(va.shape[:-1]) + [3])
    z_zero_v = np.zeros(shape)
    z_zero_v[..., :2] = va[..., :2]
    unit_v = unit_vectors(z_zero_v)  # also checks that z_zero_v is not zero vector
    x = unit_v[..., 0]
    y = unit_v[..., 1]  # ignore z component
    # todo: handle cases where x == y == 0
    restore = np.seterr(all = 'ignore')
    radians = np.where(
        np.abs(x) >= np.abs(y),
        np.where(x < 0.0, maths.pi * 3.0 / 2.0 - np.arctan(y / x), maths.pi / 2.0 - np.arctan(y / x)),
        np.where(y < 0.0, maths.pi + np.arctan(x / y), np.arctan(x / y)))
    radians = radians % (2.0 * maths.pi)
    np.seterr(**restore)
    return np.degrees(radians)


def inclination(v):
    """Returns the inclination in degrees of v (angle relative to +ve z axis)."""
    assert len(v) == 3
    unit_v = unit_vector(v)
    radians = maths.acos(unit_v[2])
    return degrees_from_radians(radians)


def inclinations(a):
    """Returns the inclination in degrees of each vector in a (angle relative to +ve z axis)."""
    assert a.ndim > 1 and a.shape[-1] == 3
    unit_vs = unit_vectors(a)
    radians = np.arccos(unit_vs[..., 2])
    return degrees_from_radians(radians)


def nan_inclinations(a):
    """Returns the inclination in degrees of each vector in a (angle relative to +ve z axis), allowing NaNs."""
    assert a.ndim > 1 and a.shape[-1] == 3
    unit_vs = nan_unit_vectors(a)
    restore = np.seterr(all = 'ignore')
    radians = np.where(np.isnan(unit_vs[..., 2]), np.nan, np.arccos(unit_vs[..., 2]))
    result = np.where(np.isnan(radians), np.nan, degrees_from_radians(radians))
    np.seterr(**restore)
    return result


def points_direction_vector(a, axis):
    """Returns an average direction vector based on first and last non-NaN points or slices in given axis."""

    assert a.ndim > 1 and 0 <= axis < a.ndim - 1 and a.shape[-1] > 1 and a.shape[axis] > 1
    if np.all(np.isnan(a)):
        return None
    start = 0
    start_slicing = [slice(None)] * a.ndim
    while True:
        start_slicing[axis] = slice(start)
        if not np.all(np.isnan(a[tuple(start_slicing)])):
            break
        start += 1
    finish = a.shape[axis] - 1
    finish_slicing = [slice(None)] * a.ndim
    while True:
        finish_slicing[axis] = slice(finish)
        if not np.all(np.isnan(a[tuple(finish_slicing)])):
            break
        finish += 1
    if start >= finish:
        return None
    if a.ndim > 2:
        mean_axes = tuple(range(a.ndim - 1))
        start_p = np.nanmean(a[tuple(start_slicing)], axis = mean_axes)
        finish_p = np.nanmean(a[tuple(finish_slicing)], axis = mean_axes)
    else:
        start_p = a[start]
        finish_p = a[finish]

    return finish_p - start_p


def dot_product(a, b):
    """Returns the dot product (scalar product) of the two vectors."""
    return np.dot(a, b)


def dot_products(a, b):
    """Returns the dot products of pairs of vectors; last axis covers element of a vector."""
    return np.sum(a * b, axis = -1)


def cross_product(a, b):
    """Returns the cross product (vector product) of the two vectors."""
    return np.cross(a, b)


def naive_length(v):
    """Returns the length of the vector assuming consistent units."""
    return maths.sqrt(dot_product(v, v))


def naive_lengths(v):
    """Returns the lengths of the vectors assuming consistent units."""
    return np.sqrt(np.sum(v * v, axis = -1))


def naive_2d_length(v):
    """Returns the length of the vector projected onto xy plane, assuming consistent units."""
    return maths.sqrt(dot_product(v[0:2], v[0:2]))


def naive_2d_lengths(v):
    """Returns the lengths of the vectors projected onto xy plane, assuming consistent units."""
    v2d = v[..., :2]
    return np.sqrt(np.sum(v2d * v2d, axis = -1))


def unit_corrected_length(v, unit_conversion):
    """Returns the length of the vector v after applying the unit_conversion factors."""
    # unit_conversion might be [1.0, 1.0, 0.3048] to convert z from feet to metres, for example
    # or [3.28084, 3.28084, 1.0] to convert x and y from metres to feet
    converted = elemental_multiply(v, unit_conversion)
    return naive_length(converted)


def manhatten_distance(p1, p2):
    """Returns the Manhattan distance between two points."""
    return abs(p2[0] - p1[0]) + abs(p2[1] - p1[1]) + abs(p2[2] - p1[2])


def manhattan_distance(p1, p2):  #  alternative spelling to above
    """Returns the Manhattan distance between two points."""
    return abs(p2[0] - p1[0]) + abs(p2[1] - p1[1]) + abs(p2[2] - p1[2])


def radians_difference(a, b):
    """Returns the angle between two vectors, in radians."""

    return maths.acos(min(1.0, max(-1.0, dot_product(unit_vector(a), unit_vector(b)))))


def degrees_difference(a, b):
    """Returns the angle between two vectors, in degrees."""

    return degrees_from_radians(radians_difference(a, b))


def rotation_matrix_3d_axial(axis, angle):
    """Returns a rotation matrix which will rotate points about axis (0, 1, or 2) by angle in degrees."""

    axis_a = (axis + 1) % 3
    axis_b = (axis_a + 1) % 3
    matrix = np.zeros((3, 3))
    matrix[axis, axis] = 1.0
    radians = radians_from_degrees(angle)
    cosine = maths.cos(radians)
    sine = maths.sin(radians)
    matrix[axis_a, axis_a] = cosine
    matrix[axis_b, axis_b] = cosine
    matrix[axis_a, axis_b] = -sine  # left handed coordinate system, eg. UTM & depth
    matrix[axis_b, axis_a] = sine
    return matrix


def rotation_3d_matrix(xzy_axis_angles):
    """Returns a rotation matrix which will rotate points about 3 axes by angles in degrees."""

    matrix = np.zeros((3, 3))
    for axis in range(3):
        matrix[axis, axis] = 1.0
    for axis in range(3):
        matrix = np.dot(matrix, rotation_matrix_3d_axial(axis, xzy_axis_angles[axis]))
    return matrix


def reverse_rotation_3d_matrix(xzy_axis_angles):
    """Returns a rotation matrix which will rotate back points about 3 axes by angles in degrees."""

    return rotation_3d_matrix(xzy_axis_angles).T


def rotate_vector(rotation_matrix, vector):
    """Returns the rotated vector."""

    return np.dot(rotation_matrix, vector)


def rotate_array(rotation_matrix, a):
    """Returns a copy of array a with each vector rotated by the rotation matrix."""

    s = a.shape
    return np.matmul(rotation_matrix, a.reshape(-1, 3).T).T.reshape(s)


def rotate_xyz_array_around_z_axis(a, target_xy_vector):
    """Returns a copy of array a suitable for presenting a cross-section using the resulting x,z values.

    arguments:
       a (numpy float array of shape (..., 3)): the xyz points to be rotated
       target_xy_vector (2 (or 3) floats): a vector indicating which direction in source xy space will end up
          being mapped to the positive x axis in the returned data

    returns:
       numpy float array of same shape as a

    notes:
       if the input points of a lie in a vertical plane parallel to the target xy vector, then the resulting
       points will have constant y values; in general, a full rotation of the points is applied, so resulting
       y values will indicate distance 'into the page' for non-planar or unaligned data
    """

    target_v = np.zeros(3)
    target_v[:2] = target_xy_vector[:2]
    rotation_angle = azimuth(target_v) - 90.0
    rotation_matrix = rotation_matrix_3d_axial(2, rotation_angle)  # todo: check sign of rotation angle
    return rotate_array(rotation_matrix, a)


def unit_vector_from_azimuth_and_inclination(azimuth, inclination):
    """Returns unit vector with compass bearing of azimuth and inclination off +z axis."""

    matrix = rotation_3d_matrix((inclination, azimuth, 0.0))
    return rotate_vector(matrix, np.array((0.0, 0.0, 1.0)))


def tilt_3d_matrix(azimuth, dip):
    """Returns a 3D rotation matrix for applying a dip in a certain azimuth.

    note:
       if azimuth is compass bearing in degrees, and dip is in degrees, the resulting matrix can be used
       to rotate xyz points where x values are eastings, y values are northings and z increases downwards
    """

    matrix = rotation_matrix_3d_axial(2, -azimuth)  # will yield rotation around z axis so azimuth goes north
    matrix = np.dot(matrix, rotation_matrix_3d_axial(0, dip))  # adjust for dip
    matrix = np.dot(matrix, rotation_matrix_3d_axial(2, azimuth))  # rotate azimuth back to original
    return matrix


def tilt_points(pivot_xyz, azimuth, dip, points):
    """Modifies array of xyz points in situ to apply dip in direction of azimuth, about pivot point."""

    matrix = tilt_3d_matrix(azimuth, dip)
    points_shape = points.shape
    points[:] -= pivot_xyz
    points[:] = np.matmul(matrix, points.reshape((-1, 3)).transpose()).transpose().reshape(points_shape)
    points[:] += pivot_xyz


def project_points_onto_plane(plane_xyz, normal_vector, points):
    """Modifies array of xyz points in situ to project onto a plane defined by a point and normal vector."""

    az = azimuth(normal_vector)
    incl = inclination(normal_vector)
    tilt_points(plane_xyz, az, -incl, points)
    points[..., 2] = plane_xyz[2]
    tilt_points(plane_xyz, az, incl, points)


def perspective_vector(xyz_box, view_axis, vanishing_distance, vector):
    """Returns a version of vector with a perspective applied."""

    mid_points = np.zeros(3)
    xyz_ranges = np.zeros(3)
    result = np.zeros(3)
    for axis in range(3):
        mid_points[axis] = 0.5 * (xyz_box[0, axis] + xyz_box[1, axis])
        xyz_ranges[axis] = xyz_box[1, axis] - xyz_box[0, axis]
    factor = 1.0 - (vector[view_axis] - xyz_box[0, view_axis]) / (vanishing_distance * (xyz_ranges[view_axis]))
    result[view_axis] = vector[view_axis]
    for axis in range(3):
        if axis == view_axis:
            continue
        result[axis] = mid_points[axis] + factor * (vector[axis] - mid_points[axis])
    return result


def determinant(a, b, c):
    """Returns the determinant of the 3 x 3 matrix comprised of the 3 vectors."""

    return (a[0] * b[1] * c[2] + a[1] * b[2] * c[0] + a[2] * b[0] * c[1] - a[2] * b[1] * c[0] - a[1] * b[0] * c[2] -
            a[0] * b[2] * c[1])


def determinant_3x3(a):
    """Returns the determinant of the 3 x 3 matrix."""

    return determinant(a[0], a[1], a[2])


def clockwise(a, b, c):
    """Returns a +ve value if 2D points a,b,c are in clockwise order, 0.0 if in line, -ve for ccw.

    note:
       assumes positive y-axis is anticlockwise from positive x-axis
    """

    return (c[0] - a[0]) * (b[1] - a[1]) - ((c[1] - a[1]) * (b[0] - a[0]))


def clockwise_triangles(p, t, projection = 'xy'):
    """Returns a numpy array of +ve values where triangle points are in clockwise order, 0.0 if in line, -ve for ccw.

    arguments:
       p (numpy float array of shape (N, 2 or 3)): points in use as vertices of triangles
       t (numpy int array of shape (M, 3)): indices into first axis of p defining the triangles
       projection (string, default 'xy'): one of 'xy', 'xz' or 'yz' being the direction of projection,
          ie. which elements of the second axis of p to use; must be 'xy' if p has shape (N, 2)

    returns:
       numpy float array of shape (M,) being the +ve or -ve values indicating clockwise or anti-clockwise
       ordering of each triangle's vertices when projected onto the specified plane and viewed in the
       direction negative to positive of the omitted axis

    note:
       assumes xyz axes are left-handed (reverse the result for a right handed system)
    """

    a0, a1 = _projected_xyz_axes(projection)

    return (((p[t[:, 2], a0] - p[t[:, 0], a0]) * (p[t[:, 1], a1] - p[t[:, 0], a1])) -
            ((p[t[:, 2], a1] - p[t[:, 0], a1]) * (p[t[:, 1], a0] - p[t[:, 0], a0])))


def in_triangle(a, b, c, d):
    """Returns True if point d lies wholly within the triangle pf ccw points a, b, c, projected onto xy plane.

    note:
       a, b & c must be sorted into anti-clockwise order before calling this function
    """

    return clockwise(a, b, d) < 0.0 and clockwise(b, c, d) < 0.0 and clockwise(c, a, d) < 0.0


def in_triangle_edged(a, b, c, d):
    """Returns True if d lies within or on the boudnary of triangle of ccw points a,b,c projected onto xy plane.

    note:
       a, b & c must be sorted into anti-clockwise order before calling this function
    """

    return clockwise(a, b, d) <= 0.0 and clockwise(b, c, d) <= 0.0 and clockwise(c, a, d) <= 0.0


def points_in_triangles(p, t, da, projection = 'xy', edged = False):
    """Returns 2D numpy bool array indicating which of points da are within which triangles.

    arguments:
       p (numpy float array of shape (N, 2 or 3)): points in use as vertices of triangles
       t (numpy int array of shape (M, 3)): indices into first axis of p defining the triangles
       da (numpy float array of shape (D, 2 or 3)): points to test for
       projection (string, default 'xy'): one of 'xy', 'xz' or 'yz' being the direction of projection,
          ie. which elements of the second axis of p  and da to use; must be 'xy' if p and da have shape (N, 2)
       edged (bool, default False): if True, points lying exactly on the edge of a triangle are included
          as being in the triangle, otherwise they are excluded

    returns:
       numpy bool array of shape (M, D) indicating which points are within which triangles

    note:
       the triangles do not need to be in a consistent clockwise or anti-clockwise order
    """

    assert p.ndim == 2 and t.ndim == 2 and da.ndim == 2 and da.shape[1] == p.shape[1]
    cwt = clockwise_triangles(p, t, projection = projection)
    a0, a1 = _projected_xyz_axes(projection)
    d_count = len(da)
    d_base = len(p)
    t_count = len(t)
    pp = np.concatenate((p, da), axis = 0)
    # build little triangles using da points and two of triangle vertices
    tp = np.empty((t_count, d_count, 3, 3), dtype = int)
    tp[:, :, :, 2] = np.arange(d_count).reshape((1, -1, 1)) + d_base
    tp[:, :, 0, 0] = np.where(cwt > 0.0, t[:, 1], t[:, 0]).reshape((-1, 1))
    tp[:, :, 0, 1] = np.where(cwt > 0.0, t[:, 0], t[:, 1]).reshape((-1, 1))
    tp[:, :, 1, 0] = np.where(cwt > 0.0, t[:, 2], t[:, 1]).reshape((-1, 1))
    tp[:, :, 1, 1] = np.where(cwt > 0.0, t[:, 1], t[:, 2]).reshape((-1, 1))
    tp[:, :, 2, 0] = np.where(cwt > 0.0, t[:, 0], t[:, 2]).reshape((-1, 1))
    tp[:, :, 2, 1] = np.where(cwt > 0.0, t[:, 2], t[:, 0]).reshape((-1, 1))
    cwtd = clockwise_triangles(pp, tp.reshape((-1, 3)), projection = projection).reshape((t_count, d_count, 3))
    if edged:
        return np.all(cwtd <= 0.0, axis = 2)
    else:
        return np.all(cwtd < 0.0, axis = 2)


@njit
def point_in_polygon(x, y, polygon):
    """Calculates if a point in within a polygon in 2D.
    
    Args:
        x (float): the point's x-coordinate.
        y (float): the point's y-coordinate.
        polygon (np.ndarray): array of the polygon's vertices in 2D.
    
    Returns:
        inside (bool): True if point is within the polygon, False otherwise.
    """
    polygon_vertices = len(polygon)
    inside = False
    p2x = 0.0
    p2y = 0.0
    xints = 0.0
    p1x, p1y = polygon[0]
    for i in numba.prange(polygon_vertices + 1):
        p2x, p2y = polygon[i % polygon_vertices]
        if y > min(p1y, p2y):
            if y <= max(p1y, p2y):
                if x <= max(p1x, p2x):
                    if p1y != p2y:
                        xints = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                    if p1x == p2x or x <= xints:
                        inside = not inside
        p1x, p1y = p2x, p2y
    return inside


@njit
def points_in_polygon(points, polygon):
    """Calculates which points are within a polygon in 2D.
    
    Args:
        points (np.ndarray): array of the points in 2D.
        polygon (np.ndarray): array of the polygon's vertices in 2D.
    
    Returns:
        polygon_points (np.ndarray): boolean array of which points are within the polygon.
    """
    polygon_points = np.empty(len(points), dtype = numba.boolean)
    for point_num in numba.prange(0, len(polygon_points)):
        polygon_points[point_num] = point_in_polygon(points[point_num, 0], points[point_num, 1], polygon)
    return polygon_points


@njit(parallel = True)
def points_in_polygons(points: np.ndarray, polygons: np.ndarray) -> np.ndarray:
    """Calculates which points are within which polygons in 2D.
    
    Args:
        points (np.ndarray): array of the points in 2D.
        polygons (np.ndarray): array of each polygons' vertices in 2D.
    
    Returns:
        polygons_points (np.ndarray): boolean array of which points are within each polygon.
    """
    polygons_points = np.empty((len(polygons), len(points)), dtype = numba.boolean)
    for polygon_num in numba.prange(0, len(polygons)):
        polygons_points[polygon_num] = points_in_polygon(points, polygons[polygon_num])
    return polygons_points


def triangle_normal_vector(p3):
    """For a triangle in 3D space, defined by 3 vertex points, returns a unit vector normal to the plane of the triangle."""

    # todo: handle degenerate triangles
    return unit_vector(cross_product(p3[0] - p3[1], p3[0] - p3[2]))


@njit
def triangle_normal_vector_numba(points):
    """For a triangle in 3D space, defined by 3 vertex points, returns a unit vector normal to the plane of the triangle."""
    v = np.cross(points[0] - points[1], points[0] - points[2])
    return v / np.linalg.norm(v)


def in_circumcircle(a, b, c, d):
    """Returns True if point d lies within the circumcircle pf ccw points a, b, c, projected onto xy plane.

    note:
       a, b & c must be sorted into anti-clockwise order before calling this function
    """

    m = np.empty((3, 3))
    m[0, :2] = a[:2] - d[:2]
    m[1, :2] = b[:2] - d[:2]
    m[2, :2] = c[:2] - d[:2]
    m[:, 2] = (m[:, 0] * m[:, 0]) + (m[:, 1] * m[:, 1])
    return determinant_3x3(m) > 0.0


def point_distance_to_line_2d(p, l1, l2):
    """Ignoring any z values, returns the xy distance of point p from line passing through l1 and l2."""

    if np.all(l2[:2] == l1[:2]):
        return naive_2d_length(p[:2] - l1[:2])
    return (abs(p[0] * (l1[1] - l2[1]) + l1[0] * (l2[1] - p[1]) + l2[0] * (p[1] - l1[1])) /
            naive_2d_length(l2[:2] - l1[:2]))


def point_distance_to_line_segment_2d(p, l1, l2):
    """Ignoring any z values, returns the xy distance of point p from line segment between l1 and l2."""

    if is_obtuse_2d(l1, p, l2):
        return naive_2d_length(p[:2] - l1[:2])
    elif is_obtuse_2d(l2, p, l1):
        return naive_2d_length(p[:2] - l2[:2])
    else:
        return point_distance_to_line_2d(p, l1, l2)


def is_obtuse_2d(p, p1, p2):
    """Returns True if the angle at point p subtended by points p1 and p2, in xy plane, is greater than 90 degrees; else False."""

    return np.dot((p1[:2] - p[:2]), (p2[:2] - p[:2])) < 0.0


def isclose(a, b, tolerance = 1.0e-6):
    """Returns True if the two points are extremely close to one another (ie.

    the same point).
    """

    #   return np.all(np.isclose(a, b, atol = tolerance))
    # cheap and cheerful alternative to thorough numpy version commented out above
    return np.max(np.abs(a - b)) <= tolerance


def is_close(a, b, tolerance = 1.0e-6):
    """Returns True if the two points are extremely close to one another (ie.

    the same point).
    """

    return isclose(a, b, tolerance = tolerance)


def point_distance_sqr_to_points_projected(p, points, projection):
    """Returns an array of projected distances squared between p and points; projection is 'xy', 'xz' or 'yz'."""

    if projection == 'xy':
        d = points[..., :2] - p[:2]
    elif projection == 'xz':
        d = points[..., 0:3:2] - p[0:3:2]
    elif projection == 'yz':
        d = points[..., 1:] - p[1:]
    else:
        raise ValueError("projection must be 'xy', 'xz' or 'yz'")
    return np.sum(d * d, axis = -1)


def nearest_point_projected(p, points, projection):
    """Returns the index into points array closest to point p; projection is 'xy', 'xz' or 'yz'."""

    # note: in the case of equidistant points, the index of the 'first' point is returned

    d2 = point_distance_sqr_to_points_projected(p, points, projection)
    return np.unravel_index(np.nanargmin(d2), d2.shape)


def area_of_triangle(a, b, c):
    """Returns the area of the triangle defined by three vertices."""

    # uses Heron's formula
    la = naive_length(a - b)
    lb = naive_length(b - c)
    lc = naive_length(c - a)
    s = 0.5 * (la + lb + lc)
    return maths.sqrt(s * (s - la) * (s - lb) * (s - lc))


def area_of_triangles(p, t, xy_projection = False):
    """Returns numpy array of areas of triangles, optionally when projected onto xy plane."""

    # uses Heron's formula
    pt = p[t]
    if xy_projection:
        la = naive_2d_lengths(pt[:, 0, :] - pt[:, 1, :])
        lb = naive_2d_lengths(pt[:, 1, :] - pt[:, 2, :])
        lc = naive_2d_lengths(pt[:, 2, :] - pt[:, 0, :])
    else:
        la = naive_lengths(pt[:, 0, :] - pt[:, 1, :])
        lb = naive_lengths(pt[:, 1, :] - pt[:, 2, :])
        lc = naive_lengths(pt[:, 2, :] - pt[:, 0, :])
    s = 0.5 * (la + lb + lc)
    return np.sqrt(s * (s - la) * (s - lb) * (s - lc))


def clockwise_sorted_indices(p, b):
    """Returns a clockwise sorted numpy list of indices b into the points p.

    note:
       this function is designed for preparing a list of points defining a convex polygon when projected in
       the xy plane, starting from a subset of the unsorted points; more specifically, it assumes that the
       mean of p (over axis 0) lies within the polygon and the clockwise ordering is relative to that mean point
    """

    # note: this function currently assumes that the mean of points bp lies within the hull of bp
    # and that the points form a convex polygon from the perspective of the mean point
    assert p.ndim == 2 and len(p) >= 3
    centre = np.mean(p, axis = 0)
    hull_list = []  # list of azimuths and indices into p (axis 0)
    for i in b:
        azi = azimuth(p[i] - centre)
        hull_list.append((azi, i))
    return np.array([i for (_, i) in sorted(hull_list)], dtype = int)


def _projected_xyz_axes(projection):
    assert projection in ['xy', 'xz', 'yz'], f'invalid projection {projection}'
    a0 = 'xyz'.index(projection[0])
    a1 = 'xyz'.index(projection[1])
    return a0, a1


# end of vector_utilities module
